#==============================================================================
# package      DE_EPAGES::Presentation::API::Object::Monitor
# state        public
# base         DE_EPAGES::Core::API::Object::MonitorShared
# description  provides a facility to report status of long-running processes
#==============================================================================
package DE_EPAGES::Presentation::API::Object::Monitor;
use base DE_EPAGES::Core::API::Object::MonitorShared;

use strict;
use HTTP::Response;
use HTTP::Status qw( HTTP_FOUND );
use DE_EPAGES::Core::API::File qw(GetFileContent WriteFile MoveFile );
use DE_EPAGES::Core::API::TestParameter qw(TestObject TestString);
use DE_EPAGES::WebInterface::API::ApplicationServer qw( GetAppServer );

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Monitor = DE_EPAGES::Presentation::API::Object::Monitor->new( %Options );
#----------------------------------------------------------------------------------------
# description  Create a new monitor object, which covers the content pages and is
#               triggered at timeout and response.
#----------------------------------------------------------------------------------------
# input        %Options | <ul>
#                   <li>see <function DE_EPAGES::Core::API::Object::MonitorShared::new></li>
#                   <li>GUID - unique id, must only contain file-system-safe characters - string</li>
#                   <li>Servlet - servlet object - DE_EPAGES::WebInterface::Servlet::BaseServlet</li>
#                   </ul> | ref.hash
# return       $Monitor | Monitor object | DE_EPAGES::Presentation::API::Object::Monitor
#========================================================================================
sub new {
    my $class = shift;
    my %Options = @_;

    TestString('GUID', $Options{'GUID'});
    TestObject('Servlet', $Options{'Servlet'}, 'DE_EPAGES::WebInterface::Servlet::BaseServlet');

    my $MonitorFileRelative = "_MONITOR_/$Options{'GUID'}.html";
    my $self = $class->SUPER::new(%Options);

    $self->{'MonitorURL'} = "/WebRoot/$MonitorFileRelative";
    $self->{'MonitorFile'} = "$ENV{EPAGES_WEBROOT}/$MonitorFileRelative";

    return $self;
}

#========================================================================================
# function     start
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Monitor->start;
#----------------------------------------------------------------------------------------
# description  Starts the monitor.
#               Writes the monitor page to the file system and then redirect the Browser
#               to this file.
#========================================================================================
sub start {
    my $self = shift;

    return if $self->isStarted;

    $self->SUPER::start;

    # Don't use $Servlet->redirect, because it prevents building a
    # response page
    my $Response = HTTP::Response->new( HTTP_FOUND ); # 302
    $Response->header( 'Location', $self->{'MonitorURL'} );

    # immediately send the redirect response
    GetAppServer()->flushOutput( $Response );
    return;
}

#========================================================================================
# function     finish
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Monitor->finish;
#----------------------------------------------------------------------------------------
# description  Called by <function run> after the main processing funtion has finished
#               if a timeout occured during processing.
#               Updates the monitor status file to redirect the browser to the actual
#               result page.
#========================================================================================
sub finish {
    my $self = shift;

    $self->writeResponse($self->get('Servlet')->response, $self->get('GUID'));

    $self->_write( $self->finishedContent );
}

sub writeMonitorContent {
    my $self = shift;

    $self->_write( $self->monitorContent );
}

#========================================================================================
# function     _write
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Monitor->_write($Content);
#----------------------------------------------------------------------------------------
# description  writes the monitor file to filesystem. With the $ContentName it's possible
#               to use different attributes for save.
#----------------------------------------------------------------------------------------
# input        $Content | content of the monitor file | string
#========================================================================================
sub _write {
    my $self = shift;
    my ($Content) = @_;
    TestString('Content', $Content);

    my $GUID = $self->get('GUID');
    $Content =~ s/#GUID\b/$GUID/g;

    my $FileName = $self->{'MonitorFile'};
    my $TempFileName = $FileName.'.tmp';
    WriteFile($TempFileName, \$Content, {'create_directory'=>1, 'utf8'=>1});
    MoveFile($TempFileName, $FileName);
    return;
}

#========================================================================================
# function     readResponse
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Presentation::API::Object::Monitor->readResponse($Response, $GUID);
#----------------------------------------------------------------------------------------
# description  Updates the given Response object with Response which was saved via
#               <function writeResponse>.
#----------------------------------------------------------------------------------------
# input        $Response | HTTP::Response object | object
# input        $GUID | unique monitor identifier | string
#========================================================================================
sub readResponse {
    shift;
    my ($Response, $GUID) = @_;

    # load response from file and parse it into a response object
    my $rResponseDump = GetFileContent(_ReponseFileName($GUID), {'utf8'=>1});
    my $DumpedResponse = HTTP::Response->parse( $$rResponseDump );

    # update current response object with parsed values
    $Response->content( $DumpedResponse->content );
    $Response->header($_ => $DumpedResponse->headers->{$_}) foreach keys %{$DumpedResponse->headers};

    # add RC and RC message
    $Response->code($DumpedResponse->code);
    $Response->message($DumpedResponse->message);

    # Problems on browser reload, cron job has to delete the monitor files
    # dont delete the .response and .html file
    return;
}

#========================================================================================
# function     writeResponse
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Presentation::API::Object::Monitor->->writeResponse($Response, $GUID);
#----------------------------------------------------------------------------------------
# description  write the given Response to file system. The GUID is the identifier to
#               identify the response later via <function readResponse>.
#----------------------------------------------------------------------------------------
# input        $Response | HTTP::Response object | object
# input        $GUID | unique monitor identifier | string
#========================================================================================
sub writeResponse {
    shift;
    my ($Response, $GUID) = @_;

    # set standard headers if not set by default
    $Response->headers->{'content-type'} = 'text/html' unless defined $Response->headers->{'content-type'};
    $Response->headers->{'cache-control'} = 'no-cache' unless defined $Response->headers->{'cache-control'};

    # evaluate result and write back the content directly into the file (EPG-12487)
    my $Result =  $Response->as_string("\n");
    my $FileName = _ReponseFileName($GUID);
    WriteFile($FileName, \$Result, {'create_directory'=>1, 'utf8'=>1});
}

# returns the name of the response file, which contains the response content
# when the function is finished
sub _ReponseFileName {
    my ($GUID) = @_;
    TestString('GUID', $GUID);

    return "$ENV{EPAGES_STORES}/_MONITOR_/$GUID.response";
}

1;
