#========================================================================================
# package      DE_EPAGES::Presentation::Servlet::Permission
# base         DE_EPAGES::Presentation::Servlet::BasePermission
# state        public
#----------------------------------------------------------------------------------------
# description  This servlet is the main dispatcher for all permission-based URL events.
#               The application server function
#               <function DE_EPAGES::WebInterface::API::ApplicationServer::run>
#               Chooses the actually used servlet package based on the configuration file
#               $ENV{EPAGES_CONFIG}/Servlet.conf.
#               Then it executes the methods <function new>, <function init> and
#               <function handle> on the servlet to process an incoming request.
#----------------------------------------------------------------------------------------
# example      # create a new servlet object
#               my $Servlet = $Package->new;
#               # initialize the servlet with parameters from the Servlet.conf section
#               $Servlet->init( \%Parameters ) if $Handler->can( 'init' );
#               # create the respose from the request
#               $Servlet->handle( $Request, $Response );
#========================================================================================
package DE_EPAGES::Presentation::Servlet::Permission;
use base qw ( DE_EPAGES::Presentation::Servlet::BasePermission );

use strict;
use feature 'switch';
use Encode qw( encode encode_utf8 );
use HTML::Parser;
use HTTP::Status; # RC_OK, RC_MOVED_PERMANENTLY etc.
use Readonly;
use Scalar::Util qw ( blessed );
use URI::Escape qw( uri_escape_utf8 );
use DE_EPAGES::Core::API::Log qw ( LogContext );
use DE_EPAGES::Content::API::Constants      qw( $SHORTURL_REGEX );
use DE_EPAGES::Content::API::ShortUrl       qw( GetObjectByURI );
use DE_EPAGES::Core::API::EpagesConfig      qw( EpagesConfig );
use DE_EPAGES::Core::API::Error             qw( Error ExistsError GetError HandleError );
use DE_EPAGES::Core::API::IDN               qw( AreDomainsEqual );
use DE_EPAGES::Core::API::Log               qw( GetLog LogDebug LogError );
use DE_EPAGES::Core::API::PerlTools         qw( CheckIpFormat GenerateGUID Memoize UsePackage trim );
use DE_EPAGES::Core::API::String            qw( EncodeNonISO1Chars QuoteXML);
use DE_EPAGES::Core::API::TestParameter     qw( TestString TestObject );
use DE_EPAGES::Core::API::Url               qw( UrlEncodeValue );
use DE_EPAGES::Database::API::Connection    qw( RunOnStore GetCurrentDBHandle );
use DE_EPAGES::Database::API::Config        qw( DatabaseConnectionData );
use DE_EPAGES::Object::API::Domain          qw( Server GetUrlInfo );
use DE_EPAGES::Object::API::Template        qw( ProcessFileTemplate );
use DE_EPAGES::Object::API::Factory         qw( LoadRootObject GetRootObjectID LoadObject LoadObjectByPath LoadObjectByGUID ExistsObjectByGUID SplitObjectPath ExistsObject );
use DE_EPAGES::Object::API::Language        qw( GetCodeByLanguageID );
use DE_EPAGES::Object::API::Table::Locale   qw( GetInfoLocale );
use DE_EPAGES::Permission::API::Permission  qw( IsActionAllowed LogAction ActionByName ExistsActionByName HaveSameSite EveryoneCanDoIt );
use DE_EPAGES::Presentation::API::Constants qw( $PROCESSMONITOR ALLOW_SESSION_IN_URL UNITY_VIEW_ACTION_EXCEPTION_REGEX );
use DE_EPAGES::Presentation::API::Form;
use DE_EPAGES::Presentation::API::Object::Monitor;
use DE_EPAGES::Presentation::API::Object::CSRFPredator;
use DE_EPAGES::Presentation::API::ServletParams; # TIEHASH
use DE_EPAGES::Presentation::API::Session   qw( GetSessionAttribute SetSessionAttribute InsertSession IsSessionTimedOut);
use DE_EPAGES::Presentation::API::Table::Session   qw( GetInfoSessionBySessionID GetInfoSessionsBySessionID DeleteSession GetInfoSession );
use DE_EPAGES::Trigger::API::Trigger        qw ( TriggerHook );
use DE_EPAGES::WebInterface::API::ApplicationServer qw( GetAppServer );
use DE_EPAGES::WebInterface::API::Config    qw( WebInterfaceConfig IsCookieAllowed );
use DE_EPAGES::WebInterface::API::Cookie    qw( GetCookies SetCookie );
use DE_EPAGES::WebInterface::API::UserAgent qw(IsUserMobilePhone);

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'PermissionDenied'   => 'Access denied for user #UserName (#UserID) action #ActionClassName.#ActionName (#ActionID) on object #ObjectName (#ObjectID) ',
        'NoViewAction'       => 'Missing parameter ViewAction',
        'OBJECT_NOT_FOUND'   => 'Object #ObjectID not found',
        'PATH_NOT_FOUND'     => 'Object "#Path" not found',
        'Object_Not_Accessible' => 'The domain name (#HTTPHost) does not match the current object (#ObjectPath)',
        'SITE_IS_CLOSED'     => 'Site closed for technical reasons.',
        'SessionNoToken'     => 'Session #IntSessionID has no security token.',
        'INVALID_CSRF_TOKEN' => 'Cookie #Cookie mismatched Form #Form, reason: #Reason'
    }
}
#ERRORDEFINITION END

Readonly my $SECURITYTOKEN_REGEXP => 'sec[0-9a-zA-Z_]+';
Readonly my $TOKEN_NAME             => 'SecToken';
Readonly my $TOKEN_COOKIE_DELETE_TIMESTAMP => '1408548169'; # Wed, 20 Aug 2014 16:05:34 +0200
Readonly my $STATIC_SALT            => '8GgkxrsQU6BCrDGg1Gh5kXrPErd14tp67';
Readonly my $INACTIVE               => 0;
Readonly my $SECTOKEN_BLACKLIST_REGEX => '^JSON';


#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet = new DE_EPAGES::Presentation::Servlet::Permission;
# example      my $PermissionServlet = new DE_EPAGES::Presentation::Servlet::Permission;
#----------------------------------------------------------------------------------------
# description  Create a new servlet, set defaults on servlet like Locale,Language,
#               OutputEncodinfg.
#----------------------------------------------------------------------------------------
# return       $Servlet | blessed object | object
#========================================================================================
sub new {
    my $class = shift;

    my $self = $class->SUPER::new;
    $self->{'Defaults'} = {
        'Locale'            => 'en_US',
        'Language'          => 'en',
        'Currency'          => 'USD',
        'OutputEncoding'    => 'utf-8',
        'ContentType'       => 'text/html',
        'StoreName'         => undef,
        'MonitorFileName'   => $ENV{'EPAGES_CARTRIDGES'} .'/DE_EPAGES/Presentation/Templates/SimpleMonitor.html',
        'ExitAfterEvent'    => 0,
        'IsSessionTimedOut' => 0,
        'IsEditorModeEnabled' => 0,
    };
    $self->{$_} = undef foreach qw( Form Cookies CookiesAtPath CookiesEnabled Params Action );
    return $self;
}

#========================================================================================
# function     form
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Form = $Servlet->form;
# example      my $Form = $Servlet->form;
#----------------------------------------------------------------------------------------
# description  Returns the form which was served by the request. On first run on servlet
#               the form will be encoded.
#----------------------------------------------------------------------------------------
# return       $Form | <package DE_EPAGES::Presentation::API::Form> object | object
#========================================================================================
sub form {
    my $self = shift;
    if (not defined $self->{'Form'}) {
        $self->{'Form'} = DE_EPAGES::Presentation::API::Form->new( $self->request );
    }
    return $self->{'Form'};
}

#========================================================================================
# function     cookies
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Cookies = $Servlet->cookies;
# syntax       $Cookie= $Servlet->cookies($Name);
# syntax       $Servlet->cookies($Name, $Value);
# syntax       $Servlet->cookies($Name, $Value, $hOptions);
#----------------------------------------------------------------------------------------
# description  Without parameter returns the function cookies which was served by the
#               request. On first run on servlet the form will be encoded.<br>
#               With one parameter (name of cookie) the value of this cookie will be
#               returned.<br>
#               With two parameter (name and value of cookie) the value of this cookie will be
#               set (includes send to client and set in <function vars>).
#----------------------------------------------------------------------------------------
# input        $Name | name of cookie (optional) | string
# input        $Value | value of cookie (optional) | string
# input        $hOptions | cookie options (optional), see
#                       <function DE_EPAGES::WebInterface::API::Cookie::SetCookie>
#                       | ref.hash
# return       $Cookies | cookies key value pairs (optional) | ref.hash
# return       $Cookie  | value of given cookie name (optional) | string
#========================================================================================
sub cookies {
    my $self = shift;
    my (@Params) = @_;

    if (scalar @Params == 0) { # get hash
        $self->{'Cookies'} = defined $self->request ? GetCookies( $self->request ): {} unless defined $self->{'Cookies'};
        return $self->{'Cookies'};
    } elsif (scalar @Params == 1) { # get
        return $self->cookies->{$Params[0]};
    } else { # set
        my ($Name,$Value,$hOptions) = @Params;
        $hOptions = {} unless defined $hOptions;
        my $OldValue = $self->cookies($Name);
        if (defined $OldValue and
            defined $Value    and
            $OldValue ne $Value
            or
            (defined $OldValue xor defined $Value)) {
            $hOptions->{Path} = $self->serverVariables->{'SCRIPT_NAME'} unless defined $hOptions->{Path};
            SetCookie( $self->response, $Name => $Value, $hOptions );
        }
        $self->cookies->{$Name} = $Value;
        return;
    }
}

#========================================================================================
# function     params
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Params = $Servlet->params;
# syntax       $Value = $Servlet->params($Name);
# syntax       $Servlet->params($Name, $Value);
#----------------------------------------------------------------------------------------
# description  This function gives an central access to all parameter of the request.
#               No matter it comes as cookie, parameter, form field or set in set meantime
#               with <function vars>.
#               Without parameter returns the function all parameter which was served by the
#               request and set via the request in<function vars>.<br>
#               With one parameter (name of variable) the value will be returned.<br>
#               With two parameter (name and value of variable) the value will be set
#               (with function <function vars>).
#----------------------------------------------------------------------------------------
# input        $Name | name of variable (optional) | string
# input        $Value | value of variable (optional) | string
# return       $Params | params key value pairs (optional) | ref.hash
# return       $Value | value of given variable name (optional) | string
#========================================================================================
sub params {
    my $self = shift;
    my (@Params) = @_;

    if (scalar @Params == 0) { # get hash
        return $self->{'Params'} if defined $self->{'Params'};

        my %params;
        tie %params, 'DE_EPAGES::Presentation::API::ServletParams', $self;
        $self->{'Params'} = \%params;
        return $self->{'Params'};
    } elsif (scalar @Params == 1) { # get
        return $self->params->{$Params[0]};
    } elsif (scalar @Params == 2) { # set
        $self->vars($Params[0], $Params[1]);
        return;
    } else {
        Error('TOO_MANY_PARAMETER');
    }
}

#========================================================================================
# function     defaults
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Defaults = $Servlet->defaults;
# syntax       $Value= $Servlet->defaults($Name);
#----------------------------------------------------------------------------------------
# description  Without parameter returns defaults of Locale,Language,OutputEncoding.<br>
#               With one parameter (name of defaults variable) the value of this defaults variable will be
#               returned.
#----------------------------------------------------------------------------------------
# input        $Name | name of defaults variable (optional) | string
# return       $Defaults | defaults key value pairs (optional) | ref.hash
# return       $Value  | value of given defaults variable name (optional) | string
#========================================================================================
sub defaults {
    my $self = shift;
    my (@Params) = @_;

    if (scalar @Params == 0) { # get hash
        return $self->{'Defaults'};
    } elsif (scalar @Params == 1) { # get
        return $self->defaults->{$Params[0]};
    } else {
        Error('TOO_MANY_PARAMETER');
    }
}

#========================================================================================
# function     init
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->init( $hParams );
#----------------------------------------------------------------------------------------
# description  Copies the parameters ViewAction, Locale and IsEditorModeEnabled from
#               the Servlet.conf file to <function vars>.
#----------------------------------------------------------------------------------------
# input        $hParams | hash with servlet parameter | ref.hash.string
#========================================================================================
sub init {
    my $self = shift;
    my ($hParams) = @_;

    $self->SUPER::init( $hParams );
    $self->defaults->{$_} = $hParams->{lc $_} foreach qw (ViewAction Locale IsEditorModeEnabled);
    $self->vars('_init_Locale', $hParams->{'locale'});
    return;
}

#========================================================================================
# function     intSessionID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $SessionID = $Servlet->intSessionID;
#----------------------------------------------------------------------------------------
# description  Returns the internal session id of the current request. A form parameter with the
#               name "SessionID" takes precedence over a cookie with the name
#               "SessionID" (external sessionid).
#----------------------------------------------------------------------------------------
# return       $SessionID | internal session id or undef if no session id was found | int
#========================================================================================
sub intSessionID {
    my $self = shift;

    my $IntSessionID = $self->vars('IntSessionID');
    return $IntSessionID if defined $IntSessionID;

    # Recover session cookie from form parameter; this is used during the transition
    # from http to https on user login.
    my $SessionID = $self->form->value('SessionID');
    return undef if defined $SessionID and $SessionID eq '';

    # Set a new SessionID cookie if session was passed as form parameter
    $self->_setSessionCookie($SessionID) if $SessionID;

    $SessionID = $self->cookies('SessionID') unless defined $SessionID;
    $SessionID = undef if defined $SessionID and $SessionID !~ /\S/;
    if (defined $SessionID) {
        my $Sessions = GetInfoSessionsBySessionID($SessionID);
        if (scalar @$Sessions) {
            my $hSession = $Sessions->[0];
            $IntSessionID = $hSession->{'IntSessionID'};
            if (IsSessionTimedOut($IntSessionID) || !defined $hSession->{'Token'}) {
                $self->vars('IsSessionTimedOut', 1);
            } else {
                $self->vars('IntSessionID', $IntSessionID);
                return $IntSessionID;
            }
        }
        # delete timedout or invalid session cookie ( possible with logout and browser.history.back )
        $self->cookies('SessionID', undef);
        return undef;
    }
    else {
        # login user by cookie if possible
        my $GUID = $self->cookies('UserGUID');
        return $self->_loginByUserGUID( $GUID ) if $GUID;
    }
}

#========================================================================================
# function     _loginByUserGUID
# state        private
#----------------------------------------------------------------------------------------
# syntax       $SessionID = $Servlet->_loginByUserGUID($GUID);
#----------------------------------------------------------------------------------------
# description  If a user with the GUID $GUID exists and login by cookie is enabled for
#               the site and the user, then the user is logged in automatically and a
#               new session id returned.
#----------------------------------------------------------------------------------------
# input        $GUID | GUID of a user | int
# return       $SessionID | internal session id or undef if user is not logged in | int
#========================================================================================
sub _loginByUserGUID {
    my $self = shift;
    my ($GUID) = @_;
    TestString( 'GUID', $GUID );

    my $Site = $self->object->getSite;
    my $HasSystemSSLCertificate = LoadRootObject()->get('HasSSLCertificate');
    my $DomainName = $Site->get('DomainName');

    if( defined $DomainName && $HasSystemSSLCertificate && !$Site->get('HasSSLCertificate') ) {
        GetLog->debug("Can't login user by cookie if the site has a domain ($DomainName) but no SSL certificate");
        return;
    }

    if (ExistsObjectByGUID($GUID)) {
        my $User = LoadObjectByGUID($GUID);

        if ( $User->get('IsLoginCookieAllowed')
          && $User->getSite()->get('LoginCookie')
          && $User->get('IsActivated')
         ) {
            # redirect user to https if possible
            if ( $HasSystemSSLCertificate && !$self->serverVariables->{'SERVER_PORT_SECURE'} ) {
                my $Location = $Site->get('WebServerScriptUrlSSL');
                my $Query = $self->request->uri->query;
                $Location .= '/?'.$Query if $Query;
                $self->redirect( $Location );
                return;
            }
            return $self->registerSession($User,1);
        } else {
            $self->cookies('UserGUID', undef);
        }
    } else {
        $self->cookies('UserGUID', undef);
    }

    return
}

#========================================================================================
# function     logout
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->logout($KeepUIToken);
#----------------------------------------------------------------------------------------
# description  Invalidates the current session of the logged-on user.
#               Deletes the session from the database and clears the cookies SessionID,
#               UserGUID, session_id and session_id.sig.
#----------------------------------------------------------------------------------------
# input        $KeepUIToken | do not delete epages ui tokens session_id and session_id.sig | boolean
#========================================================================================
sub logout {
    my $self = shift;
    my ($KeepUIToken) = @_;

    my $IntSessionID = $self->intSessionID;
    return unless $IntSessionID;

    DeleteSession( $IntSessionID );

    $self->cookies('SessionID', undef);
    $self->cookies('UserGUID', undef);
    $self->cookies('MerchantSF', undef);
    $self->vars('IntSessionID', undef);
    $self->vars('SessionID', '');

    return if $KeepUIToken;

    # delete tokens for epages_ui directly
    SetCookie($self->response, 'session_id', undef, {'Path' => '/'});
    SetCookie($self->response, 'session_id.sig', undef, {'Path' => '/'});

    return;
}

#========================================================================================
# function     redirect
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->redirect( $Location );
# syntax       $Servlet->redirect( $Location, $StatusCode );
#----------------------------------------------------------------------------------------
# description  Redirects the browser to another URL.
#----------------------------------------------------------------------------------------
# input        $Location | target URL | string
# input        $StatusCode | (optional, default: 301) HTTP status code | int
#========================================================================================
sub redirect {
    my $self = shift;
    my( $Location, $StatusCode ) = @_;

    $self->SUPER::redirect( $Location, $StatusCode );
    $self->vars('ExitAfterEvent', 1); # don't process any templates
    return;
}

#========================================================================================
# function     monitor
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Monitor = $Servlet->monitor;
#----------------------------------------------------------------------------------------
# description  Returns the monitor object.
#----------------------------------------------------------------------------------------
# return       $Monitor | instance of <package DE_EPAGES::Presentation::API::Object::Monitor> | object
#========================================================================================
sub monitor {
    my $self = shift;
    return $self->{'Monitor'};
}

#========================================================================================
# function     cookiesEnabled
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Enabled = $Servlet->cookiesEnabled;
#----------------------------------------------------------------------------------------
# description  Returns true if the client browser accepts cookies.
#               Checks if the HTTP Header "Cookie:" exists in the request.
#----------------------------------------------------------------------------------------
# return       $Enabled | true if the browser accepts cookies | boolean
#========================================================================================
sub cookiesEnabled {
    my $self = shift;
    return $self->{'CookiesEnabled'};
}

#========================================================================================
# function     cookiesAtPath
# state        public
#----------------------------------------------------------------------------------------
# syntax       $CookiesAtPath = $Servlet->cookiesAtPath;
#----------------------------------------------------------------------------------------
# description  Returns true if the session state is appended to the PATH_INFO part of
#               the URL, for example if the browser does not accept cookies.
#----------------------------------------------------------------------------------------
# return       $CookiesAtPath | true if session state is added to the URL | boolean
#========================================================================================
sub cookiesAtPath {
    my $self = shift;
    return $self->{'CookiesAtPath'};
}

#========================================================================================
# function     processMonitorContent
# state        public
#----------------------------------------------------------------------------------------
# syntax       $MonitorContentSub = $Servlet->processMonitorContent
#----------------------------------------------------------------------------------------
# description  Returns the functions to generate the content of the monitor page and
#               the finish page.
#               Returns undef if no monitor page shall be used.
#----------------------------------------------------------------------------------------
# return       $hMonitorContent | see <fuction DE_EPAGES::Core::API::Object::MonitorShared::new>
#                   <ul><li>MonitorContentSub - sub to create the content of the monitor page - string</li>
#                       <li>FinishedContentSub - sub to create the content of the finish page - string</li></ul>
#                   | ref.hash
#========================================================================================
sub processMonitorContent {
    my $self = shift;

    my $FileName = $self->params('MonitorFileName');
    my $Language = $self->params('Language');
    my $ContentSub = sub {
        my ($IsFinished) = @_;

        return ProcessFileTemplate($FileName, $Language, {
            INPUT => $self->params,
            IsContentResponse => $IsFinished,
        });
    };

    # EPG-33973: pre-process monitor page content to prevent cache corruption.
    # This is not necessary for the finish page, because it will be created after main processing.
    my $MonitorContent = $ContentSub->(0);
    return {
        MonitorContentSub => sub { return $MonitorContent },
        FinishedContentSub => sub { return $ContentSub->(1) },
    };
}

use constant DEFAULT_MONITOR_TIMEOUT => 15;

*_MonitorTimeout = Memoize( sub {
    return WebInterfaceConfig()->get( 'GLOBAL', 'MONITOR_TIMEOUT', DEFAULT_MONITOR_TIMEOUT );
});

#========================================================================================
# function     csrfPredator
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Predator = $Servlet->csrfPredator();
#----------------------------------------------------------------------------------------
# description  Returns a CSRFPredator object with request specific secrets.
#----------------------------------------------------------------------------------------
# return       $Predator | instance of <package DE_EPAGES::Permission::API::Object::CSRFPredator> | object
#========================================================================================
sub csrfPredator {
    my $self = shift;

    unless (defined $self->{'csrfPredator'}) {
        my $ServerSecret = $self->connectionCfgValue('key') // $STATIC_SALT;
        my $UserSecret = $self->storeName() . $self->site()->id;

        $self->{'csrfPredator'} = DE_EPAGES::Presentation::API::Object::CSRFPredator->new($UserSecret, $ServerSecret);
    }

    return $self->{'csrfPredator'};
}

#========================================================================================
# function     connectionCfgValue
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Value = $self->connectionCfgValue($Key);
#----------------------------------------------------------------------------------------
# description  Get information about current database connection by using
#               <function DE_EPAGES::Database::API::Config::DatabaseConnectionData>
#----------------------------------------------------------------------------------------
# input        $Key | name of key value pair | string
# return       $Value | return valu from config | string/integer
#========================================================================================
sub connectionCfgValue {
    my $self  = shift;
    my ($Key) = @_;
    TestString('Key', $Key);
    my $hSection = DatabaseConnectionData($self->storeName);
    return $hSection->{$Key};
}

#========================================================================================
# function     handleRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->handleRequest;
#----------------------------------------------------------------------------------------
# description  Handles the request and produces a response.
#               Connect to the database and executes ChangeActions and the ViewAction.
#========================================================================================
sub handleRequest {
    my $self = shift;

    return unless $self->checkDatabaseRedirect();

    my $StoreName = $self->storeName;

    my $Log = GetLog();

    # connect to database
    RunOnStore( 'Store' => $StoreName, 'Sub' => sub {

        $self->response->header('X-Store', $StoreName);
        # Disable Chrome XSS Mode Block (AD-7640)
        $self->response->header( 'X-XSS-Protection', '1' ) if $self->_isAdministrativRequest;

        $Log->debug('connected to store ' . $StoreName);

        my $hMonitorContent = $self->processMonitorContent;
        if( $PROCESSMONITOR && $hMonitorContent ) {

            my $Monitor = DE_EPAGES::Presentation::API::Object::Monitor->new(
                %$hMonitorContent,
                Servlet => $self,
                GUID => $StoreName.'-'.GenerateGUID(),
            );
            local $self->{'Monitor'} = $Monitor;
            $Monitor->run( sub { $self->_handleRequestOnStore }, _MonitorTimeout() );
        }
        else {
            local $self->{'Monitor'} = My::FakeMonitor->new;
            $self->_handleRequestOnStore;
        }
    });

    $Log->debug('disconnected');
    return;
}

sub _handleRequestOnStore {
    my $self = shift;

    eval {
        # Set cookies at path info before $self->intSessionID.
        my $Cookies = $self->requestHeaders('Cookie');
        $self->{'CookiesEnabled'} = defined $Cookies ? 1 : 0;

        # Set a test cookie at the very beginning of the session (first page) to
        # determine if the user has cookies activated on his/her browser.
        if (!defined $Cookies) {
            $self->cookies('ShopInit', 1, {'Comment' => 'To test if browser accepts cookies'});
        }

        # Extract cookies from URL path
        my $PATH_INFO = $self->serverVariables->{'PATH_INFO'}; # dont use $self->params, because it reads cookies internally
        my $CookiesAtPath = 0;
        if ($PATH_INFO =~ /;(.*)/) {
            my $PathCookies = $1;
            my %PathInfo = map { split (/=/, $_) } split (/;\s*/, $PathCookies);
            $self->cookies($_, $PathInfo{$_}) foreach keys %PathInfo;
            $CookiesAtPath = 1;
        }

        # Store these cookies and set the 'CookiesAtPath' flag
        $self->{'CookiesAtPath'} = $CookiesAtPath;

        # Init/reset sectoken for Double Cookie Submission
        $self->_bootstrapSecTokenCookie();

        my $Site = $self->site();

        # Redirect Unity
        if ($self->_isUnityRequest()) {
            $self->redirect($Site->get('WebUrlUnitySSL'));
        }

        # Redirect MobileSF
        my $UseMobileSF =  $Site->useMobileView();

        if( $self->_isMobileRequest ){
            if (!$UseMobileSF) {
                $self->_redirectRequest('mobile','sf');
                return;
            }
            $self->vars( 'IsMobile', !$self->_isClassicView );
        }
        else {
            # Redirect to .mobile insteaf of .sf if mobile device was detected
            if ( IsUserMobilePhone( $self->requestHeaders('User-Agent') ) && $UseMobileSF ) {
                return if $self->_redirectRequest('sf','mobile');
            }
            $self->vars( 'IsMobile', 0);
        }

        my $IntSessionID = $self->intSessionID; # build internal sessionid on vars
        return if !defined $IntSessionID && $self->response->code == RC_MOVED_PERMANENTLY; # redirected to site with ssl if user recognized by cookie

        my $LocaleID = $self->locale; # set locale
        $self->form->setlocale($LocaleID);

        # set variables Language and LanguageID for compatibility
        my $LanguageID = $self->languageID;
        local $DE_EPAGES::Object::API::Language::LANGUAGEID = $LanguageID;

        $self->currencyID; # get currency from request

        if (defined $IntSessionID and $self->_doSessionIntegrityTests()) {
            $self->testSessionIntegrity($IntSessionID); # throws SESSION_WRONG_ADDR, SESSION_WRONG_UA
        }

        Error('SITE_IS_CLOSED') if $Site->get('IsClosed');
        $self->setErrorCode503( 900 ) if $Site->isTemporaryClosed(); ## no critic qw( ProhibitMagicNumbers)

        my $XForwardedHostHeader = $self->requestHeaders('X-Forwarded-Host');
        if($XForwardedHostHeader) {
            my $RequestHost = $self->serverVariables->{HTTP_HOST};
            my ($LastHost) = trim((split(/,/, $XForwardedHostHeader))[-1]);
            if($RequestHost ne $LastHost) {
                $self->redirect($self->requestUrl);
                return;
            }
        }

        $self->doActions;
    };
    if( ExistsError() ) {
        my $Error = GetError();
        my $hCGIVars = $self->serverVariables;
        my %ErrorVars = map { ( $_ => $hCGIVars->{$_} ) }
            grep { exists $hCGIVars->{$_} }
            qw( SCRIPT_NAME PATH_INFO QUERY_STRING HTTP_HOST HTTP_REFERER HTTP_USER_AGENT REMOTE_ADDR );
        $Error->addVars( \%ErrorVars );
        $Error->addVars({ 'StoreName' => GetCurrentDBHandle()->cacheKey });

        given( $Error->code ) {
            when([qw( NoSearch NoSort PATH_NOT_FOUND OBJECT_NOT_FOUND ActionNotFound )]) {
                $self->handleObjectNotFound( $Error );
            }
            when('SITE_IS_CLOSED') {
                $self->handleSiteIsClosed( $Error );
            }
            when([qw(FalseToken NoToken SESSION_WRONG_ADDR SESSION_WRONG_UA)]) {
                $self->logUserFault( $Error );
                $self->handlePermissionDenied( $Error );
            }
            when('WrongURLParameter' ) {
                $self->handleWrongUrlParameter( $Error );
            }
            when('FALSE_URL_ENCODING' ) {
                $self->handleEncodingFormError( $Error );
            }
            when ('INVALID_CSRF_TOKEN') {
                GetLog->debug('invalid token ' . $Error->code);
                $self->handleError( $Error );
            }
            default {
                $self->handleError( $Error );
            }
        }
    }
}

sub _doSessionIntegrityTests {
    return EpagesConfig->get('Security', 'UseSessionIntegrityTesting', 0);
}

sub _isAdministrativRequest {
    my $self = shift;
    my $Extension = $self->serverVariables('SCRIPT_NAME') =~ /\.(\w+)$/ ? $1 : '';
    my $Root = LoadRootObject();
    my $hInfo = $Root->get([qw(URLAdminExtension URLDiagnosticsExtension)]);  # admin, diagnostics

    if ($Extension =~ m/$hInfo->{'URLAdminExtension'}|$hInfo->{'URLDiagnosticsExtension'}/x) {
        return 1;
    } else {
        return 0;
    }
}

sub _isDiagnosticsRequest {
    my $self = shift;
    my $Extension = $self->serverVariables('SCRIPT_NAME') =~ /\.(\w+)$/ ? $1 : '';

    return LoadRootObject->get('URLDiagnosticsExtension') =~ m/$Extension/x ? 1 : 0;
}

# writes a new security token cookie
sub _bootstrapSecTokenCookie {
    my $self = shift;

    my $Shop = $self->site();
    my $Domain = $Shop->get('DomainName');
    my $Host = $self->serverVariables('HTTP_HOST');

    if ($self->useSecTokenCookie) {
        # get the token via cookie
        my $TokenCookie = $self->cookies($TOKEN_NAME);

        # get the token via get/post param
        my $TokenForm = $self->form->value($TOKEN_NAME);

        if (not $TokenForm and $self->serverVariables('QUERY_STRING') =~ /$TOKEN_NAME=([^&]+)/) {
            $TokenForm = $1;
        }

        if (defined $TokenForm and $TokenForm =~ /\S/) {
            # used for SSL switch (shop domain and SSL domain differs)
            # write SecToken from form param as new cookie on the provider domain

            $self->csrfPredator->validateSecurityToken($TokenForm);
            $self->cookies($TOKEN_NAME, $TokenForm);
        }
        elsif (defined $TokenCookie) {
            # reset old cookies (older than $TOKEN_COOKIE_DELETE_TIMESTAMP)
            # used for live "patching"
            my (undef, $TimeStamp) = split /_/, $TokenCookie;

            if (defined $TimeStamp and $TimeStamp <= $TOKEN_COOKIE_DELETE_TIMESTAMP) {
                $self->cookies($TOKEN_NAME, undef);
            }
            # else: we keep the current cookie (this should be the standard case)
        }

        # write a new cookie
        unless (defined $self->cookies($TOKEN_NAME)) {
            # relax security constraints for shops with own domain
            # for shops with an own domain and a basket button on the start page
            if ($Domain and $Host eq $Domain) {
                $self->cookies($TOKEN_NAME, $self->csrfPredator->generateSecurityToken(), {'Path' => '/'});
            }
            else {
                $self->cookies($TOKEN_NAME, $self->csrfPredator->generateSecurityToken());
            }
        }
    }
}

#========================================================================================
# function     useSecTokenCookie
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $UseSecTokenCookie = $Servlet->useSecTokenCookie;
#----------------------------------------------------------------------------------------
# description  Determines if the security token cookie is in use, which protects against
#               Cross-Site Request Forgery attacks.
#----------------------------------------------------------------------------------------
# return       $UseSecTokenCookie | is security token cookie used | boolean
#========================================================================================
sub useSecTokenCookie {
    my $self = shift;
    return (
        $self->isStorefront() and
        $self->cookiesEnabled and
        IsCookieAllowed($TOKEN_NAME) and
        EpagesConfig->get('Security', 'UseSecTokenCookie', $INACTIVE) and
        LoadRootObject->get('UseSecTokenCookie')
    ) ? 1 : 0;
}

#========================================================================================
# returns true if a redirect was executed
sub _redirectRequest {
    my $self = shift;
    my ($From, $To) = @_;

    my $Location = $self->request->uri->path;
    if( $Location =~ s/(.+?)\.$From/$1\.$To/x ) {
        my $Query = $self->request->uri->query;
        $Location .= '?'.$Query if $Query;
        GetLog()->debug('Mobile device detected; redirecting to: ' . $Location);
        $self->response->header( 'Vary', '*');
        $self->redirect( $Location );
        return 1;
    }
    return 0;
}

#========================================================================================
sub _isMobileRequest {
    my $self = shift;
    return $self->request->uri->path =~ /\.mobile/x;
}

#========================================================================================
sub _isStoreFrontRequest {
    my $self = shift;
    return $self->request->uri->path =~ /\.(sf|preview)/x;
}

#========================================================================================
sub _isUnityRequest {
    my $self = shift;

    return (($self->_isStoreFrontRequest or $self->_isMobileRequest)
        and $self->site->class->alias eq 'Shop'
        and $self->site->get('IsUnityRegistered')
        and not $self->_isEp6HandledUnityRequest
    );
}

#========================================================================================
sub _isEp6HandledUnityRequest {
    my $self = shift;
    my $hValues = { 'IsEp6HandledUnityRequest' => 0 };

    TriggerHook('IsEp6HandledUnityRequest', { 'Object' => $self, 'Values' => $hValues });

    DE_EPAGES::Core::API::Log::LogDebug('IsEp6HandledUnityRequest', $hValues, 'DEBUGGING');

    return $hValues->{'IsEp6HandledUnityRequest'};
}

#========================================================================================
sub _isClassicView {
    my $self = shift;

    # Set "ClassicView" cookie to prevent usage of mobile ViewActions during the request.
    my $Cookie = $self->cookies('ClassicView');
    my $IsClassicView = $self->form->queryValue('ClassicView') // $Cookie;

    if( $Cookie && !$IsClassicView ) {
        $self->cookies( 'ClassicView', undef );
    }
    elsif( !$Cookie && $IsClassicView ) {
        $self->cookies( 'ClassicView', 1, {'Comment' => 'Do dot display mobile view'} );
    }
    return $IsClassicView;
}

#========================================================================================
# function     requestUrl
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $URL = $Servlet->requestUrl;
#----------------------------------------------------------------------------------------
# description  returns current url to compare with base url.
#               url includes protocol, servername, port, scriptname, pathinfo.
#----------------------------------------------------------------------------------------
# return       $URL | request URL (without query string) | string
#========================================================================================
sub requestUrl {
    my $self = shift;

    my $sVars = $self->serverVariables;
    my $SecurePort = $sVars->{'SERVER_PORT_SECURE'} ? 1 : 0;
    my $ServerAndPort = $self->requestHeaders('Host');
    $ServerAndPort .= ':' . $sVars->{'SERVER_PORT'} if not $SecurePort and $sVars->{'SERVER_PORT'} != 80   ## no critic qw( ProhibitMagicNumbers)
                                                        or $SecurePort and $sVars->{'SERVER_PORT'} != 443; ## no critic qw( ProhibitMagicNumbers)
    my $PathInfo = uri_escape_utf8($sVars->{'PATH_INFO'});
    $PathInfo =~ s/%2F/\//ig;
    return join '', ($SecurePort ? 'https://' : 'http://'), $ServerAndPort, $sVars->{'SCRIPT_NAME'}, $PathInfo;
}

#========================================================================================
# function     addBaseLineURL
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->addBaseLineURL;
#----------------------------------------------------------------------------------------
# description  adds base to response if necessary. if found head tag then the html tag
#               base will be appended to the head tag, otherwise the obsolate Content-Base
#               http header tag is added.
#========================================================================================
sub addBaseLineURL {
    my $self = shift;

    # set cookies at path info, set base of script
    # try to save cookies, if not possible cookies saved at url path
    #   Request Path    Response    URL
    # 1  0       0       0        not script change
    # 2  0       0       1        script with cookies at path
    # 3  0       1       0        script with cookies at path
    # 4  0       1       1        script with cookies at path
    # 5  1       0       0        not script change
    # 6  1       0       1        not script change
    # 7  1       1       0        remove path vars from script + set path vars as cookies
    # 8  1       1       1        remove path vars from script + set path vars as cookies (unless already set)
    my $Response = $self->response;
    my $ContentLocation = $self->serverVariables('SCRIPT_NAME').uri_escape_utf8($self->serverVariables('PATH_INFO'));
    $ContentLocation =~ s/%2F/\//ig;
    my $StorefrontExt = LoadRootObject()->get('URLStorefrontExtension');
    my $BaseLine = $self->vars('BaseLineContentLocation');
    if ($self->serverVariables('SCRIPT_NAME') =~ /\.($StorefrontExt|storefront)$/x) {
        $BaseLine = $self->baseLineContentLocation(undef, $self->object);
    }
    return unless defined $BaseLine;

    my $hCGIVars = $self->serverVariables;
    my $Path = ($hCGIVars->{'SERVER_PORT_SECURE'} ? 'https://' : 'http://')
        . $self->requestHeaders('Host')
        . $hCGIVars->{'SCRIPT_NAME'} . $hCGIVars->{'PATH_INFO'};
    return if $BaseLine eq $Path;

    # $BaseLineContentLocation must be encoded, because $Response->content is binary
    my $Encoding = $self->params('OutputEncoding');
    # Quote full path info to prevent XSS attacks (EPG_14348)
    my $BaseLineContentLocation = encode( $Encoding, QuoteXML( $BaseLine ) );
    my $Content = $Response->content;

    if ($Content =~ s/(<head>)/$1<base href="$BaseLineContentLocation" \/>/i ) {
        $Response->content($Content);
    }
    else {
        $Response->header( 'Content-Base' => encode('utf-8', $BaseLine));
    }
}

#========================================================================================
# function     baseLineContentLocation
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $BaseLineContentLocation = $Servlet->baseLineContentLocation($UseSSL, $Object);
#----------------------------------------------------------------------------------------
# description  Builds the base url. This url includes the path info. The path info is
#               used to save cookies at the url if cookies are not enabled or never send
#               before.
#----------------------------------------------------------------------------------------
# input        $UseSSL | url is for ssl (optional) | object
# input        $Object | object to switch context (https/http) (optional) | object
# return       $BaseLineContentLocation | url (eg 'http://localhost/epages/Store.storefront/;SessionID=429D7D210002A446273DAC14142B0603' ) | string
#========================================================================================
sub baseLineContentLocation {
    my $self = shift;
    my ($UseSSL, $Object) = @_;
    my $hCGIVars = $self->serverVariables;
    $UseSSL = $hCGIVars->{'SERVER_PORT_SECURE'} unless defined $UseSSL;
    my $Site = defined $Object ? $Object->getSite() : undef;

    my $BaseLine = defined $Site
        ? $Site->get($UseSSL ? 'WebServerScriptUrlSSL' : 'WebServerScriptUrl')
        : ($UseSSL ? 'https://' : 'http://') . $self->requestHeaders('Host') . $hCGIVars->{'SCRIPT_NAME'};

    $BaseLine = $self->mobileURL( $BaseLine ) if $self->_isMobileRequest;

    my $CookieString = '';
    my $Response = $self->response;

    if (defined $Response && ALLOW_SESSION_IN_URL) {
        my @ResponseCookies = $Response->header( 'Set-Cookie' );

        if ( not $self->cookiesEnabled and ($self->cookiesAtPath or scalar @ResponseCookies)) {
            my $hCookies = $self->cookies;
            my @Keys = sort
                grep { $_ ne 'Locale' && $_ ne 'ShopInit' }
                grep { defined $hCookies->{$_} and $hCookies->{$_} =~ /\S/ }
                keys %$hCookies;
            if( scalar @Keys ) {
                $CookieString = join( '', map { ';'.$_.'='.uri_escape_utf8($hCookies->{$_}) } @Keys );
                $CookieString =~ s/%2F/\//ig;
            }
        }
    }

    return join '/', $BaseLine, $self->createSecurityToken, $CookieString;
}

#========================================================================================
# function     mobileURL
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $MobileURL = $Servlet->mobileURL( $StorefronURL );
#----------------------------------------------------------------------------------------
# description  Converts a given storefront URL to the corresponding URL for mobile
#               devices by replacing ".sf" by ".mobile". Protocol, host name and query
#               parameters remain unchanged.
#----------------------------------------------------------------------------------------
# input        $StorefronURL | storefront URL | string
# return       $MobileURL | mobile URL | string
#========================================================================================
sub mobileURL {
    my $self = shift;
    my ($StorefronURL) = @_;

    my $Uri = URI->new($StorefronURL);
    my $Path = $Uri->path;
    $Path =~ s/\.sf/\.mobile/x;
    $Uri->path( $Path );
    return $Uri->as_string;
}

#========================================================================================
# function     createSecurityToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $TokenOrLocale = $Servlet->createSecurityToken;
#----------------------------------------------------------------------------------------
# description  Builds the security token or locale for url. The security Token starts
#               always with 'sec'
#----------------------------------------------------------------------------------------
# return       $TokenOrLocale | url part (eg 'en_US' or 'secjyQOO3ejO.w') | string
#========================================================================================
sub createSecurityToken {
    my $self = shift;
    my $IntSessionID = $self->intSessionID;
    return $self->locale unless defined $IntSessionID;

    my $Token = GetInfoSession($IntSessionID)->{'Token'};
    Error('SessionNoToken', { 'IntSessionID' => $IntSessionID }) unless defined $Token;

    return 'sec' . $Token;
}

#========================================================================================
# function     verifySecurityToken
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->verifySecurityToken;
#----------------------------------------------------------------------------------------
# description  Reads security token or locale from url. No error if token correct or no
#               user logged in.
#========================================================================================
sub verifySecurityToken {
    my $self = shift;

    return if $self->vars('TokenIsVerified');

    # if no session id then dont test the token (session timeout)
    return unless defined $self->intSessionID;

    my $PathInfo = $self->params('PATH_INFO');
    if( $PathInfo =~ m,/($SECURITYTOKEN_REGEXP)/,x ) {
        my $SecToken = $1;
        if( $SecToken ne $self->createSecurityToken ) {
            GetLog()->info(
                sprintf 'Invalid token: %s, required token: %s, intSessionID: %d, PATH_INFO: %s',
                $SecToken, $self->createSecurityToken, $self->intSessionID, $PathInfo
            );
            Error('FalseToken', {'PATH_INFO'=> $PathInfo, 'Token' => $SecToken});
        }
    }
    else {
        if( defined $self->user ) {
            GetLog()->info(
                sprintf 'Missing security token (required: %s), intSessionID: %d, PATH_INFO: %s',
                $self->createSecurityToken, $self->intSessionID, $PathInfo
            );
            Error('NoToken', {'PATH_INFO'=> $PathInfo});
        }
    }

    $self->vars('TokenIsVerified', 1);
}

#========================================================================================
# function     languageID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LanguageID = $Servlet->languageID;
#----------------------------------------------------------------------------------------
# description  Returns the language id of the current request. The id is determined from
#               language that corresponds to the current locale. See <function locale>.
#----------------------------------------------------------------------------------------
# return       $LanguageID | integer language id (default is english) | int
#========================================================================================
sub languageID {
    my $self = shift;

    # get language from locale
    my $LocaleID = $self->locale;
    return GetInfoLocale($LocaleID)->{LanguageID};
}

#========================================================================================
# function     currencyID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $CurrencyID = $Servlet->currencyID;
# syntax       $CurrencyID = $Servlet->currencyID($Reset);
#----------------------------------------------------------------------------------------
# description  Returns the currency id of the current request. The id is determined from
#               the request parameters, user preferences, cookies or site preferences.
#               If no currency was found, the default is used.
#----------------------------------------------------------------------------------------
# input        $Reset    | reset local saved values (eg: login of user) | boolean
# return       $CurrencyID | currency id (e.g. 'EUR', default='USD') | char(3)
#========================================================================================
sub currencyID {
    my $self = shift;
    my ($Reset) = @_;

    # get cached currency
    my $CurrencyID = $self->vars('Currency');
    return $CurrencyID if defined $CurrencyID and not $Reset;

    # determine locale from other parameters and cache the result
    $CurrencyID = $self->_currencyID();
    $self->vars('Currency', $CurrencyID);

    return $CurrencyID;
}

#========================================================================================
# function     setCurrencyID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->setCurrencyID($CurrencyID);
# example      $Servlet->setCurrencyID('USD');
#----------------------------------------------------------------------------------------
# description  Sets the currency for the current user.
#               Creates a new anonymous user if there is currently no user.
#----------------------------------------------------------------------------------------
# input        $CurrencyID | locale id (e.g. 'USD') | string
#========================================================================================
sub setCurrencyID {
    my $self = shift;
    my ($CurrencyID) = @_;
    TestString( 'CurrencyID', $CurrencyID );

    my $Site = $self->object->getSite();
    my $User = $self->user;
    $User = $self->createAnonymousUser($Site) unless defined $User;

    if($User->getSite()->id == $Site->id &&
       $User->get('CurrencyID') ne $CurrencyID ) {
        # correct user currency if in the same site as object
        GetLog()->debug(sprintf 'Set User.CurrencyID=%s', $CurrencyID);

        my $OldCurrencyID = $User->get('CurrencyID');

        my $hParams = {'Servlet' =>$self,
                       'NewCurrencyID' =>$CurrencyID,
                       'OldCurrencyID' => $OldCurrencyID,
                       'User' => $User,
                      };

        TriggerHook('BeforeChangeCurrency', $hParams);
        return unless $hParams->{'NewCurrencyID'};
        $User->set({ 'CurrencyID' => $hParams->{'NewCurrencyID'} });
        TriggerHook('AfterChangeCurrency', $hParams);

    }
    return;
}

# find the preferred currency
# 1. URL parameter Currency
# 2. URL parameter Locale
# 3. user
# 4. locale
# 5. site
# 6. any site currency
# 7. fixed default (USD)
sub _currencyID {
    my $self = shift;

    my $Site = $self->object->getSite();
    my @ActiveCurrencies = grep { not exists $_->{'IsActivated'} or $_->{'IsActivated'} } @{$Site->get('Currencies')};

    # get currency from parameter Currency in form
    my $CurrencyID = $self->form->value('Currency');

    if (defined $CurrencyID) { # look if currency possible
        $CurrencyID = undef unless (grep { $_->{'CurrencyID'} eq $CurrencyID } @ActiveCurrencies);
        if (defined $CurrencyID) {
            GetLog()->debug(sprintf 'Set currency %s from parameter Currency', $CurrencyID);
            $self->setCurrencyID($CurrencyID);
            return $CurrencyID;
        }
    }

    # get currency from parameter Locale in form
    my $LocaleID = $self->form->value('Locale');
    if( defined $LocaleID && $Site->existsSiteLocale($LocaleID) ) {
        my $LocaleCurrencyID = $Site->siteLocale($LocaleID)->get('Locale')->{'CurrencyID'};
        if( defined $LocaleCurrencyID && grep { $_->{'CurrencyID'} eq $LocaleCurrencyID } @ActiveCurrencies ) {
            GetLog()->debug(sprintf 'Set currency %s from parameter Locale %s', $LocaleCurrencyID, $LocaleID);
            return $LocaleCurrencyID;
        }
    }

    # get currency from user
    my $User = $self->user;
    if (defined $User) {
        $CurrencyID = $User->get('CurrencyID');
        if( defined $CurrencyID and grep { $_->{'CurrencyID'} eq $CurrencyID } @ActiveCurrencies ) {
            my $Log = GetLog();
            $Log->debug(sprintf 'Set currency %s from user defaults (%s)', $CurrencyID, $User->get('Path')) if $Log->is_debug;
            return $CurrencyID;
        }
    }

    # get currency from locale in url
    $LocaleID = $self->locale;
    if( $Site->existsSiteLocale($LocaleID) ) {
        my $LocaleCurrencyID = $Site->siteLocale($LocaleID)->get('Locale')->{'CurrencyID'};
        if( defined $LocaleCurrencyID && grep { $_->{'CurrencyID'} eq $LocaleCurrencyID } @ActiveCurrencies ) {
            GetLog()->debug(sprintf 'Set currency %s from Locale %s', $LocaleCurrencyID, $LocaleID);
            return $LocaleCurrencyID;
        }
    }

    # get default currency from Site
    $CurrencyID = $Site->get('CurrencyID');
    return $CurrencyID if defined $CurrencyID;

    return $ActiveCurrencies[0]->{'CurrencyID'} if scalar @ActiveCurrencies;

    return $self->defaults('Currency'); # default from servlet
}

#========================================================================================
# function     locale
# state        public
#----------------------------------------------------------------------------------------
# syntax       $LocaleID = $Servlet->locale;
# syntax       $LocaleID = $Servlet->locale($Reset);
#----------------------------------------------------------------------------------------
# description  Returns the locale id of the current request. The id is determined from
#               the request parameters, Cookie or object attribute.
#               If no locales was found ,the default is used.
#----------------------------------------------------------------------------------------
# input        $Reset    | reset local saved values (eg: login of user) | boolean
# return       $LocaleID | locale id (e.g. 'de_DE', default='en_US') | string
#========================================================================================
sub locale {
    my $self = shift;
    my ($Reset) = @_;

    # get cached locale
    my $LocaleID = $self->vars('Locale');
    return $LocaleID if defined $LocaleID and not $Reset;

    # determine locale from other parameters and cache the result
    $LocaleID = $self->_locale;
    $self->setLocale($LocaleID);

    return $LocaleID;
}

#========================================================================================
# function     setLocale
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->setLocale($LocaleID);
# example      $Servlet->setLocale($LocaleID);
# syntax       $Servlet->setLocale($LocaleID, $SetUser);
# example      $Servlet->setLocale($LocaleID, 0);
#----------------------------------------------------------------------------------------
# description  Sets the locale for the current request by modifying the URL of the
#               next pages.
#               If the session is associated with a user and $SetUser is true, then the
#               user default locale is also changed.
#----------------------------------------------------------------------------------------
# input        $LocaleID | locale id (e.g. 'de_DE') | string
# input        $SetUser | should the servlet modify the user's locale settings in the
#               session, optional, default is 1 | boolean
#========================================================================================
sub setLocale {
    my $self = shift;
    my ($LocaleID, $SetUser) = @_;
    TestString( 'LocaleID', $LocaleID );
    $SetUser = 1 unless defined $SetUser;

    my $LanguageID = GetInfoLocale($LocaleID)->{LanguageID};
    $self->vars('Locale', $LocaleID);
    $self->vars('Language', GetCodeByLanguageID($LanguageID));
    $self->vars('LanguageID', $LanguageID);

    my $Site = $self->object->getSite();
    my $User = $self->user;
    if($SetUser && defined $User && $User->getSite()->id == $Site->id ) {
        my $CurrencyID = $self->currencyID(1);
        if( $User->get('LocaleID') ne $LocaleID || $User->get('CurrencyID') ne $CurrencyID ) {
            GetLog()->debug(sprintf 'Set User.LocaleID=%s, CurrencyID=%s', $LocaleID, $CurrencyID);
            my $hParams = {'Servlet' =>$self,
                'NewLocaleID' =>$LocaleID,
                'OldLocaleID' => $User->get('LocaleID'),
                'User' => $User,
            };
            TriggerHook('BeforeChangeLocale', $hParams);
            $User->set({
                'LocaleID' => $LocaleID,
            });
            TriggerHook('AfterChangeLocale', $hParams);
            # set the CurrencyID for the user and existing basket
            $self->setCurrencyID($CurrencyID);
        }
    }

    $DE_EPAGES::Object::API::Language::LANGUAGEID = $LanguageID;
    return;
}

# find the preferred locale
# 1. URL parameter Locale
# 2. user
# 3. URL path
# 4. site
# 5. any site locale
# 6. fixed default (en_US)
sub _locale {
    my $self = shift;

    my $Site = $self->object->getSite();

    # get locale from parameter Locale in form
    my $LocaleID = $self->form->value('Locale');

    if( defined $LocaleID && $Site->existsSiteLocale($LocaleID) ) {
        GetLog()->debug(sprintf 'Set locale %s from parameter Locale', $LocaleID);
        return $LocaleID;
    }

    # get locale from user
    my $User = $self->user;
    if (defined $User) {
        $LocaleID = $User->get('LocaleID');
        $LocaleID = undef unless $Site->existsSiteLocale($LocaleID);
        # correct user locale if in the same site as object
        if (not defined $LocaleID and $User->getSite()->id == $Site->id) {
            $LocaleID = $Site->get('LocaleID');
            $User->set({'LocaleID'=>$LocaleID}) if defined $LocaleID;
        }
        if( defined $LocaleID ) {
            GetLog()->debug(sprintf 'Set locale %s from user defaults', $LocaleID);
            return $LocaleID;
        }
    }

    # get locale from path (configured in Servlet.conf)
    $LocaleID = $self->vars('_init_Locale');
    if( defined $LocaleID && $Site->existsSiteLocale($LocaleID) ) {
        GetLog()->debug(sprintf 'Set locale %s from URL path', $LocaleID);
        return $LocaleID;
    }

    # get default locale from Object
    $LocaleID = $Site->get('LocaleID');
    if( defined $LocaleID ) {
        GetLog()->debug(sprintf 'Set locale %s from site defaults', $LocaleID);
        return $LocaleID;
    }

    # get any locale from object
    my $aSiteLocales = $Site->get('SiteLocales');
    return $aSiteLocales->[0]->get('LocaleID') if scalar @$aSiteLocales;

    # get servlet default
    return $self->defaults('Locale');
}

#========================================================================================
# function     user
# state        public
#----------------------------------------------------------------------------------------
# syntax       $user = $Servlet->user;
#----------------------------------------------------------------------------------------
# description  Returns the user from the current session. The user is determined from
#               the SessionID parameter.
#----------------------------------------------------------------------------------------
# return       $user | user or undef if no session id was found | object
#========================================================================================
sub user {
    my $self = shift;

    my $IntSessionID = $self->intSessionID;
    return undef unless defined $IntSessionID;

    return GetSessionAttribute($IntSessionID, 'User');
}

#========================================================================================
# function     createAnonymousUser
# state        public
#----------------------------------------------------------------------------------------
# syntax       $User = $Servlet->createAnonymousUser($Site);
#----------------------------------------------------------------------------------------
# description  Returns the new anonymous user. The user will be registered at $Site.
#----------------------------------------------------------------------------------------
# input        $Site | Site (shop/system) | object
# return       $User | User | object
#========================================================================================
sub createAnonymousUser {
    my $self = shift;
    my ($Site) = @_;

    my $IntSessionID = $self->intSessionID;
    $IntSessionID = $self->registerSession unless defined $IntSessionID;
    my $Folder = $Site->folder('AnonymousUsers');

    # delete anonymous user if already exists (but is logged off)
    if( $Folder->existsChild($IntSessionID) ) {
        $Folder->child($IntSessionID)->delete;
    }

    # create a new user object
    my $User = $Site->folder('AnonymousUsers')->insertChild('AnonymousUser', {
        'Alias'      => $IntSessionID,
        'Name'       => $IntSessionID,
        'CurrencyID' => $self->params('Currency'),
        'LocaleID'   => $self->params('Locale'),
    });
    SetSessionAttribute($IntSessionID, 'User', $User);
    return $User;
}

#========================================================================================
# function     site
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Site = $Servlet->site;
#----------------------------------------------------------------------------------------
# description  Gets the site object of the current request from the X-ePages-Site header.
#----------------------------------------------------------------------------------------
# return       $Site | site of the current request | object
#========================================================================================
sub site {
    my $self = shift;

    my $SiteGUID = $self->requestHeaders('X-ePages-Site');
    die 'Missing request header X-ePages-Site' unless defined $SiteGUID;

    return LoadObjectByGUID( $SiteGUID );
}

#========================================================================================
# function     object
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object = $Servlet->object;
#----------------------------------------------------------------------------------------
# description  Returns the object on which the current action operates.
#----------------------------------------------------------------------------------------
# return       $Object | object of the current request | object
#========================================================================================
sub object {
    my $self = shift;

    my $ObjectID = $self->vars('ObjectID');
    $ObjectID = $self->params('ObjectID') unless defined $ObjectID;
    die 'ObjectID is a not plain' if ref $ObjectID ne '';
    unless( defined $ObjectID ) {

        my $Site = $self->site;

        if( defined $self->params('ObjectPath') ) {
            my $ObjectPath = $self->params('ObjectPath');
            Error('PATH_NOT_FOUND', { Path => '' }) unless $ObjectPath;
            $ObjectID = $Site->path( $ObjectPath )->id;
        }
        elsif( defined $self->params('ChangeObjectID') ) {
            ($ObjectID) = $self->form->value('ChangeObjectID');
        }
        elsif( defined $self->params('ChangeObjectPath') ) {
            my ($ObjectPath) = $self->form->value('ChangeObjectPath');
            $ObjectID = LoadObjectByPath( $ObjectPath )->id;
        }
        elsif( defined $self->params('ViewObjectID') ) {
            ($ObjectID) = $self->form->value('ViewObjectID');
        }
        elsif( defined $self->params('ViewObjectGUID') ) {
            ($ObjectID) = LoadObjectByGUID($self->params('ViewObjectGUID'))->id;
        }
        elsif( defined $self->params('ViewObjectPath') ) {
            my ($ObjectPath) = $self->form->value('ViewObjectPath');
            $ObjectID = $Site->path( $ObjectPath )->id;
        }
        else {
            $ObjectID = $Site->id;
        }

        $self->vars( 'ObjectID', $ObjectID );
    }
    my $Exists = eval { ExistsObject($ObjectID) };
    if( ExistsError() ) {
        my $Error = GetError();
        Error('OBJECT_NOT_FOUND', { ObjectID => $ObjectID }) if $Error->code eq 'FORMAT_NOT_INTEGER';
        Error($Error);
    }
    Error('OBJECT_NOT_FOUND', { ObjectID => $ObjectID }) unless $Exists;

    return LoadObject( $ObjectID );
}

#========================================================================================
# function     checkDomain
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsValid = $Servlet->checkDomain( $Object, $DomainName );
# syntax       if (!checkDomain($Object, 'www.epages.de')) { ... }
#----------------------------------------------------------------------------------------
# description  Checks if the object is allowed to be requested via the $DomainName.
#               The root object is always accessible.
#               IP-Addresses and domain names not including a dot are always allowed.
#               In addition, each object can be requested via the domain configured for
#               it's site object or the domains configured as webserver host or webserver
#               SSL host in the technical administration.
#----------------------------------------------------------------------------------------
# input        $Object | object | boolean
# input        $DomainName | server name or ip address from the HTTP "Host" header | string
# return       $IsValid | true if the domain name is valid for the object | boolean
#========================================================================================
sub checkDomain {
    my $self = shift;
    my ($Object, $httphost) = @_;

    #--- allow domain names not containing a dot (localhost, vm31, ...)
    return 1 if ($httphost !~ /\./);

    #--- allow ips
    $httphost =~ s/^\[(.+)\]$/$1/;
    my $error = CheckIpFormat($httphost);
    return 1 unless $error;

    my $System = LoadRootObject();

    #--- system object always allowed
    return 1 if ($Object->id == $System->id);

    #--- domain name of object's site
    my $Site = $Object->getSite();
    my $domain = $Site->get('DomainName');
    if (defined $domain) {
        return 1 if ($domain !~ /\./);
        return 1 if !CheckIpFormat($domain);
        return 1 if (AreDomainsEqual($httphost, $domain));
    }
    #--- webserver name configured in TBO
    my $systemdomain = $System->get('DomainName');
    return 1 if ($systemdomain !~ /\./);
    return 1 if !CheckIpFormat($systemdomain);
    return 1 if (AreDomainsEqual($httphost, $systemdomain));
    return 1 if $ENV{'EPAGES_SSL_PROXY'} && AreDomainsEqual($httphost, $ENV{'EPAGES_SSL_PROXY'});

    if($Site->class->existsAttribute('IsUnityRegistered') && $Site->get('IsUnityRegistered')) {
        # allow subdomains in case of unity shops (constructed by the function 'Server')
        return 1 if (AreDomainsEqual($httphost, Server(GetUrlInfo($Site), 0)));
    }

    #--- other domains not allowed
    return 0;
}

#========================================================================================
# function     action
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Action = $Servlet->action;
#----------------------------------------------------------------------------------------
# description  Returns the current action, else undef.
#----------------------------------------------------------------------------------------
# return       $Action | current action | object
#========================================================================================
sub action {
    return shift->{'Action'};
}

#========================================================================================
# function     hasPermission
# state        public
#----------------------------------------------------------------------------------------
# syntax       $HasPermission = $Servlet->hasPermission( $Object, $ActionName);
# syntax       $HasPermission = $Servlet->hasPermission( $Object, $Action);
#----------------------------------------------------------------------------------------
# example      unless( $Servlet->hasPermission( $Object, 'Delete' ) ) {
#                   print "Access denied";
#               }
#----------------------------------------------------------------------------------------
# description  Returns true if the current user has permission to execute an action.
#               Loggs off the current user if the site is standalone and the user does
#               not belong to this site.
#----------------------------------------------------------------------------------------
# input        $Object | object | object
# input        $Action | action object | object
# input        $ActionName | name of action | string
# return       $HasPermission | true if permission is granted | boolean
#========================================================================================
sub hasPermission {
    my $self = shift;
    my ($Object, $Action) = @_;
    $Action  = ActionByName($Object->class, $Action, $self->vars('IsMobile')) if ref $Action eq '';
    TestObject('Object', $Object);
    TestObject('Action', $Action);

    my $User = $self->user;
    # check if user belongs to same site as object
    if(not HaveSameSite($User, $Object))
    {
        # logg off the user
        GetLog->debug( sprintf( 'User %s belongs to a different site than object %s',
            $User->get('Path'), $Object->get('Path') ) );
        $self->logout;
        $User = undef;
    }

    # test permissions
    return IsActionAllowed( $Object, $User, $Action );
}

#========================================================================================
# function     testPermission
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->testPermission( $Object, $Action);
# syntax       $Servlet->testPermission( $Object, $ActionName);
#----------------------------------------------------------------------------------------
# example      eval { $Servlet->testPermission( $Object, 'Delete' ); }
#               if( ExistsError() && GetError()->code eq 'PermissionDenied' ) {
#                   print "Access denied";
#               }
#----------------------------------------------------------------------------------------
# description  Verifies if the current user has permission to execute a selected
#               Event. Throws an error if permission is denied.
#----------------------------------------------------------------------------------------
# input        $Object | object | object
# input        $Action | action object | object
# input        $ActionName | name of action | string
#========================================================================================
sub testPermission {
    my $self = shift;
    my ($Object, $Action) = @_;
    $Action  = ActionByName($Object->class, $Action, $self->vars('IsMobile')) if ref $Action eq '';
    return if $self->hasPermission( $Object, $Action );

    # throw the PermissionDenied error
    my $User = $self->user;
    my $UserID = defined $User ? $User->id : '(none)';
    my $UserName = defined $User ? $User->alias : 'Anonymous';
    Error( 'PermissionDenied', {
        'UserID'            => $UserID,
        'UserName'          => $UserName,
        'ObjectID'          => $Object->id,
        'ObjectName'        => $Object->alias,
        'ActionID'          => $Action->id,
        'ActionName'        => $Action->alias,
        'ActionClassName'   => $Action->get('ActionClass')->alias,
    });
}

#========================================================================================
# function     doAction
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Result = $self->doAction( $Action )
# syntax       $Result = $self->doAction( $ActionName )
# example      $Result = $self->doAction( 'DeleteCompany' );
#----------------------------------------------------------------------------------------
# description  Executes an event handler if the current user has sufficient permissions.
#----------------------------------------------------------------------------------------
# input        $ActionName | Name of an URLAction | string
# input        $Action | action object | object
# return       $Result | whatever the event handler returns | ?
#========================================================================================
sub doAction {
    my $self = shift;
    my ($Action) = @_;

    # get the object, default to the root object
    my $Object = $self->object;

    if( ref $Action ) {
        TestObject('Action', $Action, 'DE_EPAGES::Permission::API::Object::Action');
    }
    else {
        TestString('ActionName', $Action);
        # get the action by name
        $Action = ActionByName($Object->class, $Action, $self->vars('IsMobile'));
    }


    LogContext($self->_createContextHash($Action, $Object), sub{
        #--- check domain; object is only accessable via site's domain or domain configured in TBO
        #    of course localhost and ip addresses are always allowed
        my $httphost = $self->requestHeaders('Host');
        $httphost =~ s/:\d*$//; # remove port

        Error('Object_Not_Accessible', {'HTTPHost' => $httphost, 'ObjectPath' => $Object->get('Path')})
            unless($self->checkDomain($Object, $httphost));


        # test permission or error
        $self->testPermission( $Object, $Action );
        local $self->{'Action'} = $Action;

        my $Package = $Action->get('Package');
        my $Sub = $Action->get('FunctionName');
        die 'Action '.$Action->get('Path').' has no function name' unless defined $Sub;

        GetLog->debug('call event handler '.$Package.'::'.$Sub);
        UsePackage($Package);

        # log the action to the AUDITING log
        LogAction( 'Execute', $Object, $self->user, $Action, 'ACCEPT=' . ($self->params('ACCEPT') ? '1' : '0') );

        # execute action handler
        my $hAddVars = $Package->new(Servlet=>$self)->$Sub( $self );
        if (ref $hAddVars eq 'HASH') {
            my $hParams = $self->vars;
            $hParams->{$_} = $hAddVars->{$_} foreach keys %$hAddVars;
        } elsif (ref $hAddVars ne '' and defined blessed ($hAddVars) and $hAddVars->can('id')) {
            $self->params('ObjectID', $hAddVars->id);
        }
        GetLog->debug('return from event handler '.$Package.'::'.$Sub);
    });

    return;
}

sub _createContextHash {
    my $self = shift;
    my ($Action, $Object) = @_;
    my $ContextHash = {};
    eval {
        my $UserPath;
        my $User = $self->user();
        if ($User) {
            $UserPath = $User->pathString();
        } else {
            $UserPath = 'undef';
        }
        $ContextHash = {
            'ActionType' => $Action->class()->alias(),
            'ActionPath' => $Action->pathString(),
            'ObjectPath' => $Object->pathString(),
            'UserPath'   => $UserPath,
            'CurrencyID' => $self->currencyID(),
            'LocaleID'   => $self->locale(),
            'LanguageID' => $self->languageID(),
            'SitePath'   => $Object->getSite()->pathString(),
        };
    };
    if ( ExistsError() ) {
        my $Error = GetError();
        LogError($Error);
    }
    return $ContextHash;
}
#========================================================================================
# function     catchFormError
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->catchFormError( $cSub );
#----------------------------------------------------------------------------------------
# example       $Servlet->catchFormError( sub {
#                    $Servlet->form->executeFormError( ... );
#                });
#----------------------------------------------------------------------------------------
# description   Catch triggering of form errors. This allows you to add necessary
#                TLE variables for displaying the user input form without having to
#                specify a separate ErrorAction.
#                For use within ViewAction without any ChangeActions only!
#----------------------------------------------------------------------------------------
# input        $cSub | code to run | ref.code
#========================================================================================
sub catchFormError { ## no critic qw(RequireCommentForExportedFunctions) # perl-critic does not recognize nested subs
    my $self = shift;
    my ($cSub) = @_;

    eval {
        $cSub->();
    };
    HandleError( OnError => sub {
        my ($Error) = @_;
        if( $Error->code eq 'INVALID_FORM' ) {
            $self->_addFormErrorVars();
            return 'FIXED';
        }
    });

    return;
}

#========================================================================================
# function     _overwriteAtError
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_overwriteAtError($Error, $ErrorAction, $ErrorObjectID);
#----------------------------------------------------------------------------------------
# description  overwrites error action and object for change or view events.
#========================================================================================
sub _overwriteAtError { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Error, $ErrorAction, $ErrorObjectID) = @_;

    # get error object and action
    $ErrorObjectID = $Error->vars->{'ErrorObjectID'} unless defined $ErrorObjectID;
    $ErrorObjectID = $self->params('ErrorObjectID')  unless defined $ErrorObjectID;
    $ErrorObjectID = $self->params('ObjectID')       unless defined $ErrorObjectID;
    $self->vars('ViewObjectID', $ErrorObjectID);

    $ErrorAction = $Error->vars->{'ViewAction'} unless defined $ErrorAction;
    $ErrorAction = $self->params('ViewAction')  unless defined $ErrorAction;
    $self->vars('ViewAction', $ErrorAction);

    my $hViewVars = $Error->vars->{'ViewVars'};

    $self->_addFormErrorVars();
    if (defined $hViewVars) {
        $self->vars($_, $hViewVars->{$_}) foreach keys %$hViewVars;
    }
    return;
}

#========================================================================================
# function     doChangeActions
# state        public
#----------------------------------------------------------------------------------------

# syntax       $self->doChangeActions;
#----------------------------------------------------------------------------------------
# description  Executes all ChangeActions in the order of their appearance in the form
#               or URL.
#               Access permissions are checked before running the event handler.
#========================================================================================
sub doChangeActions {
    my $self = shift;

    my @ChangeActions   = $self->form->value('ChangeAction');
    return unless scalar @ChangeActions;

    my $ObjectID = $self->object->id;

    my @ChangeObjectIDs = $self->form->value('ChangeObjectID');
    push @ChangeObjectIDs, map { LoadObjectByPath($_)->id } $self->form->value('ChangeObjectPath');
    if (!scalar(@ChangeObjectIDs) && $ObjectID) {
        push @ChangeObjectIDs, $ObjectID;
    }

    my @ErrorActions    = $self->form->value('ErrorAction');
    my @ErrorObjectIDs  = $self->form->value('ErrorObjectID');
    push @ErrorObjectIDs, map { LoadObjectByPath($_)->id } $self->form->value('ErrorObjectPath');
    if (!scalar(@ErrorObjectIDs) && $ObjectID) {
        push @ErrorObjectIDs, $ObjectID;
    }

    # do change actions
    $self->vars('TokenIsVerified', 0);
    $self->vars('FormError', 0);
    $self->vars('FormErrors', undef);

    while (scalar @ChangeActions) {
        my $ChangeAction   = shift @ChangeActions;
        my $ChangeObjectID = shift @ChangeObjectIDs;
        my $ErrorAction    = scalar @ErrorActions ? shift @ErrorActions : undef;
        my $ErrorObjectID  = shift @ErrorObjectIDs;
        next if $ChangeAction !~ /\S/;
        $self->params('ObjectID', $ChangeObjectID) if defined $ChangeObjectID;

        $self->_checkSecToken($ChangeAction);

        unless( $self->vars('TokenIsVerified') || EveryoneCanDoIt($self->object, $ChangeAction) ) {
            $self->verifySecurityToken;
        }

        eval { $self->doAction( $ChangeAction ); };
        if( ExistsError() ) {
            my $Error = GetError();
            # ignore non INVALID_FORM errors
            if ($Error->code eq 'INVALID_FORM'){
                $self->handleInvalidForm($Error, 1, $ErrorAction, $ErrorObjectID); # 1=OnChangeAction
                last;
            }
            elsif ($Error->code eq 'PermissionDenied') {
                $self->handlePermissionDenied( $Error );
                last;
            }
            elsif ($Error->code eq 'SSL_REQUIRED_FOR_LOGIN') {
                $self->logUserFault($Error);
                $self->response->code(RC_FORBIDDEN);
                $self->viewError($Error);
                last;
            }
            else {
                Error($Error)
            }
        }
    }
}

#========================================================================================
# CSRF protection
# Okay if cookie and form param $TOKEN_NAME are equal and valid
# Okay if no cookie and no form param exists (cookies are disabled)
# Throw error 'INVALID_CSRF_TOKEN', if someone manipulated the cookie or form param
#========================================================================================
sub _checkSecToken {
    my $self = shift;
    my ($ChangeAction) = @_;
    my $Log = GetLog();

    my $ChangeActionObject = ActionByName($self->object->class, $ChangeAction, $self->vars('IsMobile'));

    if ($self->useSecTokenCookie()
        and $ChangeActionObject->get('UseSecTokenCookie') ? 1 : 0
        and not $self->isEditorMode()
        and not $ChangeAction =~ m/$SECTOKEN_BLACKLIST_REGEX/
    ) {
        my $TokenForm = $self->form->value($TOKEN_NAME);

        if (not $TokenForm and $self->serverVariables('QUERY_STRING') =~ /$TOKEN_NAME=([^&]+)/) {
            $TokenForm = $1;
        }

        my ($Hash, $TimeStamp) = split /_/, $TokenForm;

        unless (defined $Hash and defined $TimeStamp) {
            Error('INVALID_CSRF_TOKEN', {
                Token    => $TokenForm,
                Reason   => 'Token format is not acceptable'
            });
        }

        my $TokenCookie = $self->cookies($TOKEN_NAME);

        if (defined $TokenCookie and defined $TokenForm) {
            $self->csrfPredator->validateSecurityToken($TokenCookie);

            if ($TokenCookie ne $TokenForm) {
                Error('INVALID_CSRF_TOKEN', {
                    Cookie  => $TokenCookie,
                    Form    => $TokenForm,
                    Reason  => 'Tokens are not equal'
                });
            }
        }
        elsif (defined $TokenForm and not defined $TokenCookie) {
            # We treat this as a manipulation, cookie missing
            Error('INVALID_CSRF_TOKEN', {
                Cookie => $TokenCookie,
                Form   => $TokenForm,
                Reason => 'Missing form parameter or missing token'
            });
        }
        else {
            $Log->debug('No security token transmitted.');
        }
    }
    else {
        $Log->debug(
            'CSRF protection not activated. Check ' .
            WebInterfaceConfig()->fileName() .
            ' and database config ' .
            $self->storeName . '.conf'
        );
    }
}

sub isStorefront {
    my $self = shift;

    my $StorefrontExtension = LoadRootObject->get('URLStorefrontExtension');
    my $MobileStorefrontExtension = LoadRootObject->get('URLMobileStorefrontExtension');
    my $PreviewExtension = LoadRootObject->get('URLPreviewExtension');
    my $Extension = $self->serverVariables('SCRIPT_NAME') =~ /\.(\w+)$/ ? $1 : '';

    return (
        $Extension eq $StorefrontExtension or
        $Extension eq $MobileStorefrontExtension or
        $Extension eq $PreviewExtension
    ) ? 1 : 0;
}


#========================================================================================
# function     writeContentOutput
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->writeContentOutput($Content);
#----------------------------------------------------------------------------------------
# description  Fills the HTTP response with the given content in the correct encoding.
#               Hint: you can set the output encoding and content-type via the servlet
#               parameters, eg:
#               $Servlet->vars('OutputEncoding', 'utf-8');
#               $Servlet->vars('ContentType', 'text/x-json');
#----------------------------------------------------------------------------------------
# input        $Content | your content | string
#========================================================================================
sub writeContentOutput { ## no critic qw(RequireCommentForExportedFunctions)
    my $self = shift;
    my ($Content, $ContentType) = @_; # parameter $ContentType is for compatibility with ePages 6.0.0-6.0.2
    $ContentType = $self->params('ContentType') unless defined $ContentType;

    # replace  by ss for German in Switzerland
    if ($ContentType =~ /^text/ && $self->locale eq 'de_CH') {
        $Content =~ s/\N{U+00DF}/ss/g;
    }

    # encode output
    my $Encoding = $self->params('OutputEncoding');
    if( $Encoding eq 'utf-8' ) {
        $Content = encode_utf8( $Content );
    }
    else {
        Encode::_utf8_on($Content); ##no critic (ProtectPrivateSubs)
        if(uc($Encoding) eq 'ISO-8859-1') {
            $Content = EncodeNonISO1Chars($Content);
        }
        $Content = encode( $Encoding, $Content );
    }
    GetLog->debug('content encoded');

    # make output
    $self->response->headers->content_type( "$ContentType; charset=$Encoding" );
    $self->response->headers->header('Content-Language', $self->params('Language'));
    $self->response->content( $Content );
    if ($ContentType eq 'text/html' && defined GetCurrentDBHandle() && !$self->params('ExitAfterEvent')) {
        $self->addBaseLineURL();
    }
    $self->vars('ResponseContentProcessed', 1);
    return;
}

#========================================================================================
# function     handleInvalidForm
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->handleInvalidForm($Error, $OnChangeAction, $ErrorAction, $ErrorObjectID);
#----------------------------------------------------------------------------------------
# description  calls <function doViewAction> for the given error.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
# input        $OnChangeAction | call on change or view action (used for some tests) | boolean
# input        $ErrorAction | error action (optional)| object
# input        $ErrorObjectID | error object identifier (optional)| integer
#========================================================================================
sub handleInvalidForm {
    my $self = shift;
    my ($Error, $OnChangeAction, $ErrorAction, $ErrorObjectID) = @_;

    if($OnChangeAction) {
        $self->_overwriteAtError($Error, $ErrorAction, $ErrorObjectID);
    } else {
        # remove change actions from error actions
        my @ChangeActions   = $self->form->value('ChangeAction');
        my @ErrorActions    = $self->form->value('ErrorAction');
        my @ErrorObjectIDs  = $self->form->value('ErrorObjectID');

        foreach my $ChangeAction (@ChangeActions) {
            shift @ErrorActions;
            shift @ErrorObjectIDs;
        }
        my $ObjectID = $self->viewObject()->id;
        my $ViewAction = $self->params('ViewAction');
        $ErrorAction = shift @ErrorActions;
        $ErrorObjectID = shift @ErrorObjectIDs;

        $self->_overwriteAtError($Error, $ErrorAction, $ErrorObjectID);
        if( $ObjectID == $self->object->id and $ViewAction eq $self->vars('ViewAction') ) {
            Error('NO_ERROR_ACTION', {}, $Error);
        }
    }

    $self->doViewAction;
    return;
}

#========================================================================================
# function     _addFormErrorVars
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->_addFormErrorVars;
#----------------------------------------------------------------------------------------
# description  Assigns FormError variables from Form to Servlet.
#========================================================================================
sub _addFormErrorVars {
    my $self = shift;

    $self->vars('FormError', 1);
    $self->vars('FormErrors', $self->form->errors);
    $self->vars('FormRealErrors', $self->form->formErrors);
    $self->form->resetFormErrors;
}

#========================================================================================
# function     doViewAction
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->doViewAction;
#----------------------------------------------------------------------------------------
# description  Executes a ViewAction and fills the HTTP response using the template and
#               TLE variables provided by the event handler function.
#               Access permissions are checked before running the event handler.
#========================================================================================
sub doViewAction {
    my $self = shift;
    GetLog->debug('start doViewAction');

    # get view action and object
    my $ViewAction = $self->params('ViewAction') or Error( 'NoViewAction' );

    # get the object, default to the root object
    my $Object = $self->viewObject;

    # test if we have an admin request and a unity shop, then rewrite viewaction to Unity-$Action
    my $Site = $Object->getSite;
    if ($Site->class->alias eq 'Shop') {
        my $IsUnity = ($Site->class->existsAttribute('IsUnityRegistered') and $Site->get('IsUnityRegistered')) ? 1 : 0;

        # throw error 'action not found' in case of a Unity View action in a non Unity shop
        Error('ActionNotFound') if (!$IsUnity && $ViewAction =~ m/^Unity/);

        if ($self->_isAdministrativRequest and not $self->_isDiagnosticsRequest) {

            if ($IsUnity and not $ViewAction =~ UNITY_VIEW_ACTION_EXCEPTION_REGEX) {
                # add prefix Unity to name if not exists
                $ViewAction = "Unity$ViewAction";
                $self->vars('ViewAction', $ViewAction);
            }
        }
    }

    # get the action by name
    my $Action = ActionByName($Object->class, $ViewAction, $self->vars('IsMobile'));

    #check if action is a view action
    $self->verifySecurityToken() unless $Action->instanceOf('ViewAction');

    my $ObjectID = $self->viewObject()->id;
    $self->vars('ObjectID', $ObjectID);

    # filter out false linked images e.g. http://<servername>/epages/Store.sf/en_GB/null.gif
    if (($ObjectID == GetRootObjectID()) and ($ViewAction eq 'View' || $ViewAction eq 'MBO-ViewGeneral')  ) {
       $self->handleObjectNotFound( DE_EPAGES::Core::API::Error->new(
                                      Code => 'InvalidURL',
                                      Vars => {'URI'=> $self->request->uri}
                                    ));
       return;
    }

    # do action
    eval { $self->doAction( $ViewAction ); };
    if( ExistsError() ) {
        my $Error = GetError();
        if( $Error->code eq 'PermissionDenied' ) {
            $self->handlePermissionDenied( $Error );
            return;
        }
        elsif( $Error->code eq 'INVALID_FORM' ) {
            $self->handleInvalidForm( $Error, 0 ); # 0=OnViewAction
            return;
        }
        Error( $Error );
    }
    GetLog->debug('ViewAction end', $ViewAction );

    # ExistAfterEvent doesnt process content
    GetLog->debug('ExitAfterEvent') if $self->params('ExitAfterEvent');
    return if $self->params('ExitAfterEvent');

    # process content
    my $Content = $self->processContent;
    GetLog->debug('content processed');

    $self->writeContentOutput($Content);
    GetLog->debug('end doViewAction');
    return;
}

#========================================================================================
# function     viewObject
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $Object = $Servlet->viewObject;
#----------------------------------------------------------------------------------------
# description  Returns the Object for the ViewAction.
#----------------------------------------------------------------------------------------
# return       $Object | currently active object | DE_EPAGES::Object::API::Object::Object
#========================================================================================
sub viewObject {
    my $self = shift;

    my $ObjectID = $self->params('ViewObjectID');
    if (defined $ObjectID) {
        $self->params('ObjectID', $ObjectID);
    } else {
        if( defined $self->params('ViewObjectPath') ) {
            my $Site = $self->site;
            $ObjectID = $Site->path( $self->form->value('ViewObjectPath') )->id;
            $self->params('ViewObjectID', $ObjectID);
        } else {
            $ObjectID = $self->params('ObjectID');
            $self->params('ViewObjectID', $ObjectID);
        }
    }

    my $Object = eval { LoadObject($ObjectID) };
    if( ExistsError() ) {
        my $Error = GetError();
        Error('OBJECT_NOT_FOUND', { ObjectID => $ObjectID }) if $Error->code eq 'FORMAT_NOT_INTEGER';
        Error($Error);
    }

    return $Object;
}

#========================================================================================
# function     doActions
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->doActions;
#----------------------------------------------------------------------------------------
# description  Executes all change actions and then executes the view action.
#               Processing is terminated immediately at the first error.
#               Calls <function handleError> if an error occurs while processing an
#               action.
#========================================================================================
sub doActions {
    my $self = shift;

    $self->vars('BaseLineContentLocation', $self->baseLineContentLocation);
    $self->doChangeActions;
    $self->vars('BaseLineContentLocation', $self->baseLineContentLocation);
    $self->doViewAction unless $self->vars('ResponseContentProcessed');

    if ($self->intSessionID or $self->_isAdministrativRequest) {
        $self->response->header('Cache-Control', 'no-store');
        $self->response->header('Pragma', 'no-cache');
    }
}

#========================================================================================
# function     handlePermissionDenied
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->handlePermissionDenied( $Error );
#----------------------------------------------------------------------------------------
# description  Dispays the login form as a result of an unauthorized view action.
#               Triggers the action ViewSimpleLoginForm and sets the followg TLE variables:
#               <ul>
#                   <li>PermissionDenied = 1</li>
#                   <li>PermissionDeniedText = $Error->shortMessage</li>
#               </ul>
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
#========================================================================================
sub handlePermissionDenied {
    my $self = shift;
    my ($Error) = @_;
    TestObject( 'Error', $Error );

    $self->vars('PermissionDenied', 1);
    $self->vars('PermissionDeniedText', $Error->shortMessage);
    $self->doAction( 'ViewSimpleLoginForm' );
    $self->writeContentOutput($self->processContent);
}

#========================================================================================
# function     handleObjectNotFound
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->handleObjectNotFound( $Error );
#----------------------------------------------------------------------------------------
# description  Handles the error "Object not found" or "Path not found".
#               This function sets the HTTP status 404 - "Object not found" and displays
#               an error message and logs a warning message containg QUERY_STRING
#               and HTTP_REFERER.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
#========================================================================================
sub handleObjectNotFound {
    my $self = shift;
    my ($Error) = @_;
    TestObject( 'Error', $Error );

    $self->logUserFault( $Error );

    my $ObjectPath = $self->params('ObjectPath');
    my $ViewAction = $self->params('ViewAction');
    my $RootObject = LoadRootObject();
    my $ExistingObject;
    if (($Error->code eq 'PATH_NOT_FOUND' or ($Error->code eq 'ActionNotFound' and $ViewAction eq 'View'))
     and defined $ObjectPath and (not defined $ViewAction or $ViewAction eq 'View')) {
        my $aPathElements = SplitObjectPath($ObjectPath);
        $ExistingObject = $RootObject; shift @$aPathElements;
        # find last existing object
        foreach my $Alias (@$aPathElements) {
            last unless $Alias =~ /\S/; # EPG-4883: stop on empty elements
            last unless $ExistingObject->existsChild($Alias);
            $ExistingObject = $ExistingObject->child($Alias);
        }
        # find last object with view action = View
        # get the action by name
        while (defined $ExistingObject and not ExistsActionByName($ExistingObject->class, 'View')) {
            $ExistingObject = $ExistingObject->parent;
        }
    }

    # provide a link to a higher level category or shop homepage
    my $ViewObject = defined $ExistingObject ? $ExistingObject : $RootObject;
    if ($ViewObject->getSite()->id != $RootObject->id) {
        $self->vars('ExistingObject', $ViewObject->tleHash);
    }

    # show error page in correct language/locale for shop or system
    $self->vars('ObjectID', $ViewObject->id);
    $self->locale; # set locale
    local $DE_EPAGES::Object::API::Language::LANGUAGEID = $self->languageID;
    $self->currencyID; # get currency from request

    # send the status code 404 - File not found to remove the page from search engines
    $self->response->code( RC_NOT_FOUND );
    $self->viewError($Error);
    return;
}

#========================================================================================
# function     handleSiteIsClosed
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->handleSiteIsClosed( $Error );
#----------------------------------------------------------------------------------------
# description  Handles the error "Site is closed".
#               This function sets the HTTP status 503 - "This Web site has been
#               temporarily closed.".
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
#========================================================================================
sub handleSiteIsClosed {
    my $self = shift;
    my ($Error) = @_;
    TestObject( 'Error', $Error );

    # send the status code 503 - File not found to remove the page from search engines
    $self->setErrorCode503( 900 ); ## no critic qw( ProhibitMagicNumbers )
    $self->viewError($Error);
    return;
}

#========================================================================================
# function     handleMonitorNotFound
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->handleMonitorNotFound( $Error );
#----------------------------------------------------------------------------------------
# description  Handles the error "FILE_NOT_FOUND" for not found monitor files.
#               This function sets the HTTP status 404 - "Object not found" and displays
#               an error message and logs a warning message containg QUERY_STRING
#               and HTTP_REFERER.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
#========================================================================================
sub handleMonitorNotFound {
    my $self = shift;
    my ($Error) = @_;
    TestObject( 'Error', $Error );

    $self->logUserFault( $Error );

    # send the status code 404 - File not found to remove the page from search engines
    $self->response->code( RC_NOT_FOUND );

    $self->viewError($Error);
    return;
}

#========================================================================================
# function     handleWrongUrlParameter
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->handleWrongUrlParameter( $Error );
#----------------------------------------------------------------------------------------
# description  Handles the error "url parameter has wrong value ".
#               This function sets the HTTP status 400 - "Bad Request" and displays
#               an error message and logs a warning message containing QUERY_STRING
#               and HTTP_REFERER.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
#========================================================================================
sub handleWrongUrlParameter {
    my $self = shift;
    my ($Error) = @_;
    TestObject( 'Error', $Error );

    $self->logUserFault( $Error );

    # send the status code 400 - Client Bad Request
    $self->response->code( RC_BAD_REQUEST );

    $self->viewInvalidURLError($Error);
    return;
}

#========================================================================================
# function     handleEncodingFormError
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->handleEncodingFormError( $Error );
#----------------------------------------------------------------------------------------
# description  Handles the error "false encoding in url".
#               This function sets the HTTP status 400 - "Bad Request" and displays
#               an error message and logs a warning message containing QUERY_STRING
#               and HTTP_REFERER.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
#========================================================================================
sub handleEncodingFormError {
    my $self = shift;
    my ($Error) = @_;
    TestObject( 'Error', $Error );

    $self->logUserFault( $Error );

    # send the status code 400 - Client Bad Request
    $self->response->code( RC_BAD_REQUEST );
    $self->vars('OutputEncoding', 'utf-8');
    $self->vars('ContentType', 'text/html');
    $self->viewInvalidURLError($Error);
    return;
}

#========================================================================================
# function     handleError
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->handleError( $Error );
#----------------------------------------------------------------------------------------
# description  Handles an error. This function logs the error and then displays it
#               using the <function viewError> method.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
#========================================================================================
sub handleError {
    my $self = shift;
    my ($Error) = @_;
    TestObject('Error', $Error);


    if ($Error->code eq 'Object_Not_Accessible' or ($Error->code eq 'INVALID_DEREFERRER_TOKEN_USER_FAULT' and $self->params('ViewAction') eq 'DereferredRedirect')) {
        $self->logUserFault($Error);
    }
    else {
        LogError($Error);
    }

    $self->response->code(RC_INTERNAL_SERVER_ERROR);
    $self->viewError($Error);
}

#========================================================================================
# function     viewError
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->viewError( $Error );
#----------------------------------------------------------------------------------------
# description  Displays an error page, which is processed by <function processErrorContent>.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
#========================================================================================
sub viewError {
    my $self = shift;
    my ($Error) = @_;

    GetLog->debug('viewError: '.$Error->code);

    # make output
    $self->response->code( RC_INTERNAL_SERVER_ERROR ) if $self->response->code == RC_OK;
    $self->response->headers->header( 'Cache-Control' => 'no-cache' );
    $self->writeContentOutput( $self->processErrorContent($Error) );

    GetLog->debug('end handleError');
}

#========================================================================================
# function     viewInvalidURLError
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->viewInvalidURLError( $Error );
#----------------------------------------------------------------------------------------
# description  Displayse an invalid url error, which is processed by <function processErrorContent>.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
#========================================================================================
sub viewInvalidURLError {
    my $self = shift;
    my ($Error) = @_;

    GetLog->debug('viewInvalidURLError: '.$Error->code);

    # make output
    $self->response->code( RC_BAD_REQUEST ); # Bad Request
    $self->response->headers->header( 'Cache-Control' => 'no-cache' );

    # encode output
    my $Content = encode_utf8( $self->processErrorContent($Error) );

    # make output
    $self->response->headers->content_type( 'text/html; charset=utf-8' );
    $self->response->content( $Content );

    GetLog->debug('end viewInvalidURLError');
}

#========================================================================================
# function     processErrorContent
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Content = $Servlet->processErrorContent( $Error );
#----------------------------------------------------------------------------------------
# description  Builds an error message using the SimpleError template in English.
#----------------------------------------------------------------------------------------
# input        $Error | error object | object
# return       $Content | error message | string
#========================================================================================
sub processErrorContent {
    my $self = shift;
    my ($Error) = @_;
    TestObject( 'Error', $Error );

    my $hVars = {
        Error => {
            Code => $Error->code,
            GUID => $Error->guid,
            Hash => $Error->hash,
            ShortMessage => $Error->shortMessage,
            LongMessage => $Error->longMessage,
            Vars => $Error->vars,
        }
    };

    my $Template = "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Presentation/Templates/SimpleError.html";
    return ProcessFileTemplate( $Template, 'en', $hVars );
}

#========================================================================================
# function     isEditorMode
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsEditorMode = $Servlet->isEditorMode;
#----------------------------------------------------------------------------------------
# description  isbackoffice request (from form data)
#----------------------------------------------------------------------------------------
# return       $IsEditorMode | template should be processed with editor | boolean
#========================================================================================
sub isEditorMode {
    my $self = shift;

    # was set before
    my $IsBO = $self->vars('IsEditorMode');
    return $IsBO if defined $IsBO;

    # was set at form
    $IsBO = $self->params('IsEditorMode') && $self->defaults('IsEditorModeEnabled');
    $IsBO = $IsBO ? 1:0;
    $self->vars('IsEditorMode', $IsBO);
    return $IsBO;
}

#========================================================================================
# function     isEditorNavElements
# state        public
#----------------------------------------------------------------------------------------
# syntax       $isDesignEditor = $Servlet->isEditorNavElements;
#----------------------------------------------------------------------------------------
# description  Returns whether the "IsEditorNavElements" parameter is set, indicating
#               a preview frame with the SF in quick design / adv. design mode.
#----------------------------------------------------------------------------------------
# return       1 if in design mode, otherwise 0 | boolean
#========================================================================================
sub isEditorNavElements {
    my $self = shift;

    # Quick design / adv. design preview frame with SF
    return $self->params('IsEditorNavElements') ? 1 : 0;
}

#========================================================================================
# function     processContent
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Content = $Servlet->processContent;
#----------------------------------------------------------------------------------------
# description  process template and var hash to response string
#----------------------------------------------------------------------------------------
# return       $Content | processed template | string
#========================================================================================
sub processContent {
    my $self = shift;

    my $Object = $self->object;
    my $ViewAction = $self->params('ViewAction');
    my $Action = ActionByName($Object->class, $ViewAction, $self->vars('IsMobile'));

    my $hVars = $Object->tleHash;
    my $ServletVars = $self->vars;

    $hVars->{$_} = $ServletVars->{$_} foreach keys %$ServletVars;
    $hVars->{'INPUT'} = $self->params;
    $hVars->{'VIEWACTION'} = $Action->tleHash;
    $hVars->{'System'} = LoadRootObject()->tleHash;

    my $TemplateFile = $self->vars('TemplateFile');
    TestString('TemplateFile', $TemplateFile);

    return ProcessFileTemplate( $TemplateFile, $self->params('Language'), $hVars );
}

#========================================================================================
# function     registerSession
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IntSessionID = $Servlet->registerSession($User, $PerCookie);
#----------------------------------------------------------------------------------------
# description  creates a new session id and save this as cookie.
#               Applies the locale and currency preferences of the user ($User).
#----------------------------------------------------------------------------------------
# input        $User | user (optional) | object
# input        $PerCookie | user identified by cookie (optional) | bit
# return       $IntSessionID | processed template | string
#========================================================================================
sub registerSession {
    my $self = shift;
    my ($User, $PerCookie) = @_;
    my $SessionID = InsertSession($User, $PerCookie);

    my $IntSessionID = GetInfoSessionBySessionID($SessionID)->{'IntSessionID'};
    $self->vars('IntSessionID', $IntSessionID);
    $self->vars('TokenIsVerified', 1); # new token
    $self->vars('BaseLineContentLocation', $self->baseLineContentLocation);

    # bind UserAgent to session
    SetSessionAttribute($IntSessionID, 'UserAgent', $self->params('HTTP_USER_AGENT'));

    $self->_setSessionCookie( $SessionID );
    if( defined $User ) {
        # switch locales for user
        $self->locale(1);
        $self->currencyID(1);
    }

    return $IntSessionID;
}

#========================================================================================
# function     _setSessionCookie
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Servlet->_setSessionCookie( $SessionID );
#----------------------------------------------------------------------------------------
# description  Sets the cookie "SessionID" with HttpOnly flag and secure flag if the
#               connection is SSL-secured.
#----------------------------------------------------------------------------------------
# input        $SessionID | session id | string
#========================================================================================
sub _setSessionCookie {
    my $self = shift;
    my ($SessionID) = @_;

    my $hCGIVars = $self->serverVariables;
    my %Options = ('HttpOnly' => 1);
    $Options{'secure'} = 1 if $hCGIVars->{'SERVER_PORT_SECURE'};
    $self->cookies('SessionID', $SessionID, \%Options);
    return;
}

#========================================================================================
# function     historyUri
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->historyUri;
#----------------------------------------------------------------------------------------
# description  returns uri for history.
#----------------------------------------------------------------------------------------
# return       $HistoryURI | URL | string
#========================================================================================
sub historyUri { return shift->request->uri; }

#========================================================================================
# function     logUserFault
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->logUserFault( $Error );
#----------------------------------------------------------------------------------------
# description  Logs a short message without stack trace using the log level "WARN" and
#               log category 'USERFAULT::'.$Error->code
#----------------------------------------------------------------------------------------
# input        $Error | error object | DE_EPAGES::Core::API::Error
#========================================================================================
sub logUserFault {
    my $self = shift;
    my ($Error) = @_;
    TestObject( 'Error', $Error );

    my $Message = $Error->shortMessage;
    $Message .= "\n";
    $Message .= 'Code = ' . $Error->code . ':' . $Error->guid . "\n";
    $Message .= 'URI=' . $self->request->uri . "\n";
    my $hServerVariables = $self->serverVariables;
    foreach my $var (qw(HTTP_USER_AGENT HTTP_REFERER HTTP_HOST)) {
        my $Value = $hServerVariables->{$var};
        $Message .= "$var=$Value\n" if defined $Value && $Value ne '';
    }
    GetLog('USERFAULT::'.$Error->code)->warn( $Message );
}

#========================================================================================
# function     isSecureConnection
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->isSecureConnection();
# example      unless ($Servlet->isSecureConnection()) {
#                   # error handling, redirect, logging, ...
#               }
#----------------------------------------------------------------------------------------
# description  Determines if the request was sent on a secure port, given the store got
#               a SSL certificate.
#               Returns always 1 if there's no server SSL certificate at all.
#----------------------------------------------------------------------------------------
# return       $IsSecure | secure connection indicator | boolean
#========================================================================================
sub isSecureConnection {
    my $self = shift;

    return (LoadRootObject()->get('HasSSLCertificate') and not $self->serverVariables('SERVER_PORT_SECURE')) ? 0 : 1;
}

#========================================================================================
# function     testSecureConnection
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->testSecureConnection();
#----------------------------------------------------------------------------------------
# description  Abort the current action and displays a SSL_REQUIRED_FOR_LOGIN error page
#               with status code 403 if <function isSecureConnection> returns false.
#               For use in sensitive ChangeActions which require a secure connection.
#========================================================================================
sub testSecureConnection {
    my $self = shift;

    unless ($self->isSecureConnection()) {
        Error('SSL_REQUIRED_FOR_LOGIN');
    }
}

#========================================================================================
# function     testPostRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->testPostRequest();
#----------------------------------------------------------------------------------------
# description  Abort the current action and displays a POST_REQUIRED_FOR_REQUEST error page
#               with status code 403 if <function isPOSTRequest> returns false.
#               For use in sensitive ChangeActions which require more security.
#========================================================================================
sub testPostRequest {
    my $self = shift;

    unless ($self->isPOSTRequest()) {
        Error('POST_REQUIRED_FOR_REQUEST');
    }
}

sub _isMatchingUserAgent {
    my $self = shift;
    my ($SessionID) = @_;
    my $SessionUserAgent = GetSessionAttribute($SessionID, 'UserAgent');
    my $RequestUserAgent = $self->request->header('user-agent');

    if (defined $SessionUserAgent and $SessionUserAgent eq $RequestUserAgent) {
        return 1;
    } else {
        return 0;
    }
}

sub _isMatchingAddress {
    my $self = shift;
    my ($SessionID) = @_;
    my $SessionRemoteIP = GetSessionAttribute($SessionID, 'RemoteIP');
    my $RequestRemoteIP = $self->serverVariables('REMOTE_ADDR');

    if (defined $SessionRemoteIP and $SessionRemoteIP eq $RequestRemoteIP) {
        return 1;
    } else {
        return 0;
    }
}

#========================================================================================
# function     testSessionIntegrity
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->testSessionIntegrity($SessionID);
#----------------------------------------------------------------------------------------
# description  Check if current request matches the 'UserAgent' and 'RemoteIP' stored
#               in the current session. The later is only checked for administrative requests
#               (MBO/PBO/Diagnostics). May throw SESSION_WRONG_UA, SESSION_WRONG_ADDR
#----------------------------------------------------------------------------------------
# input        $SessionID | current session id | integer
#========================================================================================
sub testSessionIntegrity {
    my $self = shift;
    my ($SessionID) = @_;

    if($self->_isMatchingUserAgent($SessionID)) {
        if ($self->_isAdministrativRequest() and not $self->_isMatchingAddress($SessionID)) {
            my $hInfo = {Expected => GetSessionAttribute($SessionID, 'RemoteIP'), Got => $self->serverVariables('REMOTE_ADDR')};
            $self->logout();
            Error('SESSION_WRONG_ADDR', $hInfo);
        } else {
            # everthing is ok
        }
    } else {
        my $hInfo = {Expected => GetSessionAttribute($SessionID, 'UserAgent'), Got => $self->request->header('user-agent')};
        $self->logout();
        Error('SESSION_WRONG_UA', $hInfo);
    }
}

#========================================================================================
# function     redirectToSecureConnection
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->redirectToSecureConnection($DomainURL);
# example      $Servlet->redirectToSecureConnection(
#                   $Servlet->object->getSite->get('WebServerScriptUrlSSL')
#               );
#----------------------------------------------------------------------------------------
# description  Abort the current action and reloads the current GET request on a secure
#               connection using the given secure domain URL if
#               <function isSecureConnection> returns false.
#               For use in sensitive ViewActions which require a secure connection.
#----------------------------------------------------------------------------------------
# input        $DomainURL | secure URL for redirect | string
#========================================================================================
sub redirectToSecureConnection($) {
    my $self = shift;
    my ($DomainURL) = @_;

    TestString('DomainURL', $DomainURL, 1);

    unless ($self->isSecureConnection()) {
        my $Location;

        my $SiteGUID = $self->requestHeaders('X-ePages-Site');
        die 'Missing request header X-ePages-Site' unless defined $SiteGUID;

        my $Shop = LoadObjectByGUID( $SiteGUID );
        my $URI;
        if ($self->serverVariables('QUERY_STRING') =~ /^URI=([^&]+)/) {
            $URI = $1;
        }

        # Do we have a short-url?
        if(defined $URI and $URI =~ $SHORTURL_REGEX) {
            #Do we have a own ssl certificate for the shop domain?
            if($self->object->getSite->get('HasSSLCertificate')) {
                $Location = join('',
                    $self->object->getSite->get('ProtocolAndServerSSL'),
                    '/',
                    $URI
                );
            } else {
                my $ScriptName = $Shop->get('WebServerScriptName');
                my ($ContentObject, $LocaleID) = GetObjectByURI($Shop, $URI);

                # if shortURL exists then show this
                if (defined $ContentObject) {
                    $Location = $ScriptName . '/?ObjectPath=' . UrlEncodeValue($ContentObject->get('Path'));
                    $Location = $ScriptName . '/' . $LocaleID . '/?ObjectPath=' . UrlEncodeValue($ContentObject->get('Path')) if (defined $LocaleID);
                }

                if (defined $LocaleID) {
                   $Location .= '&Locale=' . $LocaleID;
                }
            }
        } else {
            $Location = join('',
                $DomainURL,
                $self->serverVariables('PATH_INFO'),
                '?',
                $self->serverVariables('QUERY_STRING')
            );
        }

        $self->redirect($Location);

        return;
    }
}

#========================================================================================
# function     forcePOSTForNonTestShops
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->forcePOSTForNonTestShops();
#----------------------------------------------------------------------------------------
# description  Checks if request is a POST request. If this check fails a form error is
#               executed. Internal test shops are not checked.
#========================================================================================
sub forcePOSTForNonTestShops {
    my $self = shift;
    my $Shop = $self->object->getSite();
    my $IsTestShop = ($Shop->className eq 'Shop' && $Shop->get('IsInternalTestShop'));

    unless ($IsTestShop or $self->isPOSTRequest()) {
        $self->form->addFormError({
            'Reason'=> 'POST_METHOD_REQUIRED',
            'Form'  => 'Login',
            'Name'  => 'Login',
        });
        $self->form->executeFormError;
    }
}

#========================================================================================
# function     isPOSTRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Servlet->isPOSTRequest();
#----------------------------------------------------------------------------------------
# description  Return true if current request was send with method POST.
# return       $IsPost | Is post request ? | boolean
#========================================================================================
sub isPOSTRequest {
    return shift->params('REQUEST_METHOD') eq 'POST';
}

# for compatibility with $Servlet->monitor->blockTimeOut;
package My::FakeMonitor; ## no critic qw(ProhibitMultiplePackages)
sub new {
    my $class = shift;
    return bless {}, $class;
}

sub blockTimeOut{
}

1;
