#==============================================================================
# package      DE_EPAGES::Product::API::WebService::ProductService14
# base         DE_EPAGES::Product::API::WebService::ProductService13
# state        public
# description  Implements the Product WebService.
#==============================================================================
package DE_EPAGES::Product::API::WebService::ProductService14;
use base DE_EPAGES::Product::API::WebService::ProductService13;

use strict;
use DE_EPAGES::XML::API::SOAPDateTimeFormatXML;
use DE_EPAGES::Core::API::File qw ( ConvertFileNameChars WriteFile );
use DE_EPAGES::Object::API::Language qw( GetCodeByLanguageID );

sub new {
    my $self = shift;
    my @Options = @_;
    $self->SUPER::new((
        'ServiceUrn' => 'urn://epages.de/WebService/ProductService/2015/01',
        'ServiceTypesUrn' => 'urn://epages.de/WebService/ProductTypes/2015/01',
        @Options,
    ));
}

#add EnergyLabelsString for update
sub _stringAttributeNames { ##no critic (ProhibitUnusedPrivateSubroutines)
    return (shift->SUPER::_stringAttributeNames,'EnergyLabelsString');
}

sub _cleanupProductUpdate { ##no critic (ProhibitUnusedPrivateSubroutines)
    my $self = shift;
    my ($product, $hInProduct) = @_;

    my $hCleanProduct = $self->SUPER::_cleanupProductUpdate($product, $hInProduct);
    $hCleanProduct->{'EnergyLabelSourceFile'} = $hInProduct->{'EnergyLabelSourceFile'} if exists $hInProduct->{'EnergyLabelSourceFile'};
    $hCleanProduct->{'ProductDataSheet'} = $hInProduct->{'ProductDataSheet'} if exists $hInProduct->{'ProductDataSheet'};

    return $hCleanProduct;
}

#========================================================================================
# function     buildSoapProduct
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->buildSoapProduct( $product, $InputTaxModel,
#                   $aAttributes, $aCurrencies, $aLanguages );
#----------------------------------------------------------------------------------------
# description  No WebService-Method!
#               Build SOAP::Data object for Product.
#----------------------------------------------------------------------------------------
# input        $product | product | Product object
# input        $InputTaxModel | input tax model | tax model id
# input        $aAttributes | list of additional attributes | array ref
# input        $aCurrencies | list of currencies | array ref
# input        $aLanguages | list of languages | array ref
# return       $SoapData  | a SOAP::Data structure representing the product | SOAP::Data
#========================================================================================
sub buildSoapProduct {
    my $self = shift;
    my ($product, $InputTaxModel, $aAttributes, $aCurrencies, $aLanguages) = @_;

    #--- get base soap object ---
    my $soapProduct = $self->SUPER::buildSoapProduct(
        $product, $InputTaxModel, $aAttributes, $aCurrencies, $aLanguages
    );

    #--- get base soap value ---
    my $hOutProduct = $soapProduct->value();

    my $PublicEnergyLabelFile;
    if ($product->get('EnergyLabelSourceFile')) {
        $PublicEnergyLabelFile = $self->soapString( $product->get('PublicPath').'/'.$product->get('EnergyLabelSourceFile'));
    }

    $hOutProduct->{'EnergyLabelsString'}    = $self->soapString( $product->get('EnergyLabelsString') );
    $hOutProduct->{'EnergyLabelSourceFile'} = $self->soapString( $product->get('EnergyLabelSourceFile') );
    $hOutProduct->{'EnergyLabelSourceFilePath'} = $PublicEnergyLabelFile;
    $hOutProduct->{'ProductDataSheet'}      = $self->soapLocList( $product, 'ProductDataSheet', $aLanguages );
    $hOutProduct->{'ProductDataSheetPath'}  = $self->_locPathList($product, 'ProductDataSheet', $aLanguages);

    #--- return soap object ---
    return $self->soapData( 'Product', $hOutProduct, 'TGetInfo_Return');
}

#========================================================================================
# function     setParameters
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->setParameters( $product, $hProduct )
#----------------------------------------------------------------------------------------
# description  No WebService-Method!
#               Sets parameters of a product object (for create and update products).
#----------------------------------------------------------------------------------------
# input        $product | product to update | Product object
# input        $hProduct | data to assign to the product
#                             <ul>
#                               <li>ExpiryDate    - datetime
#                             </ul> | ref.hash
#========================================================================================
sub setParameters {
    my $self = shift;
    my ($product, $hInProduct) = @_;

    #--- call base method
    $self->SUPER::setParameters($product, $hInProduct);

    # set more Product parameters
    my $hInProductParams = {};

    $hInProductParams->{'EnergyLabelsString'} = $hInProduct->{'EnergyLabelsString'} if defined $hInProduct->{'EnergyLabelsString'};

    if ( $hInProduct->{'EnergyLabelSourceFile'} && $hInProduct->{'EnergyLabelSourceFile'}->{'Value'}  ) {
        my $hSourceFileInfo = $hInProduct->{'EnergyLabelSourceFile'};
        if ( $hSourceFileInfo->{'FileContent'} ) {
            $hSourceFileInfo->{'Value'} = ConvertFileNameChars($hSourceFileInfo->{'Value'});
            #upload public file
            my $FileName = $product->get('PublicPath').'/'.$hSourceFileInfo->{'Value'};
            WriteFile($FileName, \$hSourceFileInfo->{'FileContent'}, { binary=>1, create_directory=>1} );
        }
        #override FileName parameter
        $hInProductParams->{'EnergyLabelSourceFile'} = $hSourceFileInfo->{'Value'};
    }

    # save attachments for product
    if ( $hInProduct->{'ProductDataSheet'} ) {
        foreach my $LocalizedDataSheet ( @{$hInProduct->{'ProductDataSheet'}} ) {
            if ( $LocalizedDataSheet->{'FileContent'}) {
                $LocalizedDataSheet->{'Value'} = ConvertFileNameChars($LocalizedDataSheet->{'Value'});
                #upload public file
                my $FileName = $product->get('PublicPath').'/'.$LocalizedDataSheet->{'Value'};
                WriteFile($FileName, \$LocalizedDataSheet->{'FileContent'}, { binary=>1, create_directory=>1} );

                #override FileName parameter
                $LocalizedDataSheet->{'FileName'} = $FileName;
            }
        }

        # set localized params
        $self->setLocalizedValues($product, 'ProductDataSheet', $hInProduct->{'ProductDataSheet'});
    }

    # set params
    $product->set( $hInProductParams ) if scalar %$hInProductParams;

}

sub _locPathList {
    my $self = shift;
    my ($object, $attributeId, $aLanguageIds) = @_;

    my $aLocValues = [];
    foreach my $langid (@$aLanguageIds) {
        push @$aLocValues, SOAP::Data->type('TLocalizedValue')->name('LocalizedValue' => {
            'LanguageCode' => $self->soapString( GetCodeByLanguageID($langid) ),
            'Value'        => $self->soapString( $object->get('PublicPath').'/'.$object->get( $attributeId, $langid )),
        });
    }
    return $self->soapList($aLocValues, 'TLocalizedValue', undef);
}

1;
