#==============================================================================
# package      DE_EPAGES::Product::API::WebService::ProductService9
# base         DE_EPAGES::Product::API::WebService::ProductService8
# state        public
# description  Implements the Product WebService.
#==============================================================================
package DE_EPAGES::Product::API::WebService::ProductService9;
use base DE_EPAGES::Product::API::WebService::ProductService8;

use strict;
use DE_EPAGES::Core::API::Log qw ( LogDebug );
use DE_EPAGES::Content::API::ShortUrl qw ( GetShortUrlByClass );
use DE_EPAGES::Core::API::File qw ( ConvertFileNameChars WriteFile );

sub new {
    my $self = shift;
    my @Options = @_;
    $self->SUPER::new((
        'ServiceUrn' => 'urn://epages.de/WebService/ProductService/2011/10',
        'ServiceTypesUrn' => 'urn://epages.de/WebService/ProductTypes/2011/10',
        @Options,
    ));
}

#========================================================================================
# function     setParameters
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->setParameters( $product, $hProduct )
#----------------------------------------------------------------------------------------
# description  No WebService-Method!
#               Sets parameters of a product object (for create and update products).
#----------------------------------------------------------------------------------------
# input        $product | product to update | Product object
# input        $hProduct | data to assign to the product
#                             <ul>
#                               <li>Name              - LocalizedString
#                               <li>Description       - LocalizedString
#                               <li>LongDescription   - LocalizedString
#                               <li>Keywords          - LocalizedString
#                               <li>TaxClass          - ObjectPath
#                               <li>IsVisible         - boolean
#                               <li>IsNew             - boolean
#                               <li>IsDailyPrice      - boolean
#                               <li>OrderUnit         - ObjectPath
#                               <li>PriceQuantity     - float
#                               <li>MinOrder          - float
#                               <li>IntervalOrder     - float
#                               <li>WeightUnit        - ObjectPath
#                               <li>Weight            - float
#                               <li>RefUnit           - ObjectPath
#                               <li>RefAmount         - integer
#                               <li>RefContentAmount  - float
#                               <li>StockLevel        - float
#                               <li>StockLevelAlert   - float
#                               <li>Manufacturer      - string
#                               <li>ManufacturerSKU   - string
#                               <li>DeliveryPeriod    - integer
#                               <li>Length            - float
#                               <li>Height            - float
#                               <li>Width             - float
#
#                               <li>DefaultSubProduct - boolean
#                               <li>HasSubOwnPrices   - boolean
#                               <li>IsDefault         - boolean
#
#                               <li>IsAvailable         - boolean
#                               <li>AvailabilityComment - LocalizedString
#                               <li>AvailabilityDate    - datetime
#
#                               <li>URI - LocalizedString
#
#                             </ul> | ref.hash
#========================================================================================
sub setParameters {
    my $self = shift;
    my ($Product, $hInProduct) = @_;
    LogDebug('setParameters:',[$Product, $hInProduct]);

    # save private attachments for download products
    if ( $hInProduct->{'IsDownloadProduct'} ) {
        foreach my $DownloadProductMap ( @{$hInProduct->{'DownloadProductMaps'}} ) {
            if ( $DownloadProductMap->{'FileContent'} & !$DownloadProductMap->{'IsExternal'}) {
                #upload private file
                my $FileName = $Product->get('PrivatePath').'/'.ConvertFileNameChars({$DownloadProductMap->{'FileName'}});
                WriteFile($FileName, \$DownloadProductMap->{'FileContent'}, { binary=>1, create_directory=>1} );

                #override FileName parameter
                $DownloadProductMap->{'FileName'} = $FileName;
            }
        }
    }

    $self->SUPER::setParameters($Product, $hInProduct);

    if (not defined $hInProduct->{'URI'}) {
        foreach my $LanguageID (@{$self->getLanguageIDs()}) {
            if ( ! $Product->get('URI', $LanguageID) ) {
                my $URI = GetShortUrlByClass(
                    $Product->get('Parent'),
                    $Product->get('Class'),
                    $Product->get('NameOrAlias', $LanguageID),
                    $LanguageID
                );
                $Product->set( {'URI'=>$URI}, $LanguageID ) if $URI;
            }
        }
    }

}

1;
