use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Core::API::Error qw( GetError ExistsError );
use DE_EPAGES::Core::API::Log qw( LogError );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Object::API::Factory qw ( LoadObject LoadObjectByPath );
use DE_EPAGES::Shop::API::System qw( GetUnityShopIDs );
use DE_EPAGES::WebInterface::API::MessageCenter qw( SynchronizeCache );

#========================================================================================
# function         PatchUnityProductZoom
# state            public
#----------------------------------------------------------------------------------------
# syntax           PatchUnityProductZoom();
#----------------------------------------------------------------------------------------
# description      Sets the default value of UseProductZoom for existing shops
#========================================================================================
sub PatchUnityProductZoom {
    my ($ShopAlias) = @_;

    if (defined $ShopAlias) {
        # single shop run
        my $Shop = LoadObjectByPath('/Shops/'.$ShopAlias);
        if ($Shop->isUnity()) {
            # Unity shop
            _PatchUnityProductZoom($Shop);
        } else {
            # non Unity shop
           print 'Shop ' . $ShopAlias . " is not a Unity shop.\n";
        }
    } else {
        # all Unity shops
        my $aShopIDs = GetUnityShopIDs();

        foreach my $ShopID (@$aShopIDs) {
            eval {
                my $Shop = LoadObject($ShopID);
                _PatchUnityProductZoom($Shop);
            };
            warn('Patch of the Unity product zoom failed for shop ID ' . $ShopID . " with error \"" . GetError() . "\".")
                if ExistsError();
        }
    }

    eval { SynchronizeCache(); };
    LogError(GetError()) if ExistsError();

    return;
}

sub _PatchUnityProductZoom {
    my ($Shop) = @_;

    # default the setting UseProductZoom to 1
    $Shop->set({ 'UseProductZoom' => 1 });

    return;
}

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $ShopAlias;
    my $Help;

    GetOptions(
        'help'        => \$Help,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'shopid=s'    => \$ShopAlias,
    );

    usage() if $Help;

    unless ( defined $StoreName ) {
        print STDERR "Missing parameter -storename.\n\n";
        usage();
    }
    my $time1 = time;

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub { PatchUnityProductZoom($ShopAlias); },
    );
    my $time2 = time;
    printf "need %.1f seconds\n", $time2 - $time1;
}

sub usage {
    print <<END_USAGE;
Description:
    Sets the default value of UseProductZoom for existing shops
Usage:
    perl $0 [options] [flags] datafiles

options:
    -passwd     database user password                          ()
    -storename  name of store                                   ()
    -shopalias  alias of a single shop                      (optional)
                (if not given, the script is preformed for all Unity shops)
flags:
    -help       show the command line options

Example:
    perl $0 -storename Store
END_USAGE
    exit 2;
}

RunScript( 'Sub' => \&Main );