#========================================================================================
# package      DE_EPAGES::SageOne::t::unit::UI_Shop_SaveSettings
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  unit test for UI::Shop
#========================================================================================
package DE_EPAGES::SageOne::t::unit::UI_Shop_SaveSettings;
use base DE_EPAGES::Test::API::UnitTest::Test;
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::SageOne::API::Constants qw ( SAGE_MAP_FIELDS_TLE );
use DE_EPAGES::Test::Mock::Mockify;
use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use DE_EPAGES::Test::Mock::MockifyServlet;
use HTTP::Request;
use Test::Exception;
use Test::More;

use strict;

sub test {
    my $self = shift;
    $self->_buildMocks();

    $self->test_SaveSettings_Successful();
    $self->test_SaveSettings_RejectedBySageOne();
    $self->test_SaveSettings_WebServiceError();

    done_testing();
}

sub test_SaveSettings_Successful {
    my $self = shift;
    my $SubTestName = (caller(0))[3];
    my $Servlet = $self->_createServlet();
    my $ShopConfigService = $self->_createShopConfigService();
    my $UI_Shop = $self->_createUIShop($ShopConfigService);

    lives_ok( sub {  $UI_Shop->SaveSettings($Servlet); }, "$SubTestName - test if SaveSettings lives");
    is($Servlet->vars('FormError'), undef, "$SubTestName - test if Form doesn't contain field with error");
}

sub test_SaveSettings_RejectedBySageOne {
    my $self = shift;
    my $SubTestName = (caller(0))[3];
    my $Servlet = $self->_createServlet();
    my $ShopConfigService = $self->_createShopConfigService_WithInvalidForm();
    my $UI_Shop = $self->_createUIShop($ShopConfigService);
    throws_ok( sub { $UI_Shop->SaveSettings($Servlet); },
                   qr/INVALID_FORM: ABC/,
                   "$SubTestName - test if SaveSettings throws INVALID_FORM"
         );
    my $Form = $self->memory()->get('Form');
    my $aExpectedForm =  [{
                           'Reason' => 'This field doesn\'t exists',
                           'ErrorPackage' => 'DE_EPAGES::SageOne::UI::Shop',
                           'Name' => 'LedgerAccountInvoice',
                           'Form' => 'SageOneConfig'
                         }];
    is_deeply($Form->formErrors(), $aExpectedForm, "$SubTestName - test if Form does contains missing field for LedgerAccountInvoice");
    is($Form->existsFormError(), 1, "$SubTestName - test if Form does contain field with error");
}

sub test_SaveSettings_WebServiceError {
    my $self = shift;
    my $SubTestName = (caller(0))[3];
    my $Servlet = $self->_createServlet();
    my $ShopConfigService = $self->_createShopConfigService_WithWebserviceError();
    my $UI_Shop = $self->_createUIShop($ShopConfigService);
    $UI_Shop->SaveSettings($Servlet);
    is($Servlet->vars('WebserviceError'), 1, "$SubTestName - test if SaveSettings thrown WebserviceError and display message error");
}
#----------------------------------------------------------------------------------------
# helpers
#----------------------------------------------------------------------------------------
sub _buildMocks {
    my $self = shift;
    my $Form = $self->_createForm();
    $self->memory()->add('Form', $Form);
}

sub _createShopConfigService {
    my $self = shift;
    my $aParameterList = [$self->_createShop()];
    my $Mockify = DE_EPAGES::Test::Mock::Mockify->new(
                   'DE_EPAGES::SageOne::API::Webservice::ShopConfigService',
                   $aParameterList
              );
    my $hContent = {'LedgerAccountInvoice' => 123};
    $Mockify->addMockWithReturnValueAndParameterCheck('updateShopConfig', undef, [{'hashref' => $hContent}]);
    return $Mockify->getMockObject();
}

sub _createShopConfigService_WithInvalidForm {
    my $self = shift;
    my $aParameterList = [$self->_createShop()];
    my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::SageOne::API::Webservice::ShopConfigService', $aParameterList);
    my $DTO = $self->_createFieldErrorDTO();
    $Mockify->addMock('updateShopConfig', sub{ Error('ShopConfigUpdateError', {'ConfigFieldError' => [$DTO]}); });
    return $Mockify->getMockObject();
}

sub _createShopConfigService_WithWebserviceError {
    my $self = shift;
    my $aParameterList = [$self->_createShop()];
    my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::SageOne::API::Webservice::ShopConfigService', $aParameterList);
    $Mockify->addMock('updateShopConfig', sub{ Error('WebserviceError'); });
    return $Mockify->getMockObject();
}

sub _createUIShop {
    my $self = shift;
    my ($ShopConfigService) = @_;

    my $Mockify = DE_EPAGES::Test::Mock::Mockify->new( 'DE_EPAGES::SageOne::UI::Shop', [] );
    $Mockify->addMockWithConditionalReturn('_BuildShopUrl', 'http://epages6returnurl.com/', [
        {'object' => 'DE_EPAGES::Shop::API::Object::Shop'},
        {'hashref' => { 'ViewAction' => 'MBO-SageOneSettings' }},
        {'hashref' => {'Type' => 'admin', 'UseSSL' => 1, 'AddToken' => 1}},
        {'object' => 'DE_EPAGES::WebInterface::Servlet::BaseServlet'},
    ]);
    $Mockify->addMockWithReturnValueAndParameterCheck('_createShopConfigService', $ShopConfigService, [{'object'=>'DE_EPAGES::Shop::API::Object::Shop'}]);
    return $Mockify->getMockObject();
}

sub _createServlet {
    my $self = shift;
    my $Shop = $self->_createShop();
    my $MockifyServlet = DE_EPAGES::Test::Mock::MockifyServlet->new($Shop);

    my $Form = $self->memory()->get('Form');
    $MockifyServlet->addMockWithReturnValue('form', $Form);
    return $MockifyServlet->getMockObject();
}

sub _createForm {
    my $self = shift;
    my $Request = HTTP::Request->new( 'POST', 'http://epages.com' );
    my $Mockify = DE_EPAGES::Test::Mock::Mockify->new('DE_EPAGES::Presentation::API::Form', [$Request]);

    my $hValues = {'ledger_account_invoice' => 123};
    $Mockify->addMockWithReturnValueAndParameterCheck('form', $hValues, [{'object' => 'DE_EPAGES::Shop::API::Object::Shop'}, {'string' => 'SageOneConfig'}]);
    $Mockify->addMock('executeFormError', sub{ Error('INVALID_FORM', {'ABC' => 123}); });
    return $Mockify->getMockObject();
}

sub _createShop {
    my $self = shift;
    return $self->mockFactory()->buildShop({
            'Alias' => 'MyShopAlias',
            'RESTAPIURL' => 'http://external/base/url/',
            'ID' => 18847,
            'GUID' => 'R34ND0MVAEWLU5634ES',
            'Site' => $self->_createSite()
        })->getMockObject();
}

sub _createFieldErrorDTO {
    my $self = shift;
    my $DTO = DE_EPAGES::SageOne::API::DTO::ShopConfigFieldError->new();
    $DTO->formFieldName(SAGE_MAP_FIELDS_TLE->{'ledger_account_invoice'});
    $DTO->name('ledger_account_invoice');
    $DTO->value('Old Value');
    $DTO->message('This field doesn\'t exists');
    return $DTO;
}

sub _createSite {
    my $self = shift;
    my $Mockify = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Object::API::Object::Site', {'ID' => 123, 'GUID' => 'R34ND0MVAEWLU5634ES',});
    return $Mockify->getMockObject();
}

__PACKAGE__->runScript();
