#========================================================================================
# package      DE_EPAGES::TLE::API::NumberFormatter
# state        public
#----------------------------------------------------------------------------------------
# description  This tle handler provides functionalities around numbers. The the perl internal locale is used.
#               <br />This handler supports the following tle variable:
#               <ul>
#                 <li>ExampleInteger - 12345
#                 <li>ExampleFloat - 12345.6789
#                 <li>SmallExampleFloat - 12.34
#               </ul>
#               This handler supports the following formats:
#               <ul>
#                 <li>[integer] integer formatted</li>
#                 <li>[float] float formatted</li>
#                 <li>[money] money formatted</li>
#                 <li>[px] pixel formatted</li>
#                 <li>[neg] negative value</li>
#                 <li>[number] natural number formatted, without separators</li>
#                 <li>[lazyinteger] strips extra characters from an integer value, for example "width: -123px" becomes "-123"</li>
#               </ul>
#               The most formatter are overwritten with <package DE_EPAGES::Object::API::TLE::SiteFormatter>.
#               <br />This formatter uses the existing functionality of <package DE_EPAGES::Core::API::Object::NumberFormat>.
# example      #ExampleInteger[px] => 12345px
# example      #ExampleInteger[integer] => 12.345 (de_DE)
# example      #ExampleFloat[float] => 12.345,68 (de_DE)
# example      #ExampleFloat[money] => 12.345,68  (de_DE)
#========================================================================================
package DE_EPAGES::TLE::API::NumberFormatter;

use strict;
use DE_EPAGES::Core::API::Object::NumberFormat;

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler = new DE_EPAGES::TLE::API::NumberFormatter;
# example      my $NumberFormatter  = new DE_EPAGES::TLE::API::NumberFormatter;
#----------------------------------------------------------------------------------------
# description  Creates a new handler.
#----------------------------------------------------------------------------------------
# return       $Handler | blessed object | object
#========================================================================================
sub new { return bless {}, shift; }

#========================================================================================
# function     register
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler->register( $Processor );
#----------------------------------------------------------------------------------------
# description  This function registers this handle on the processor.
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
#========================================================================================
sub register {
    my $self = shift;
    my ($Processor) = @_;
    $Processor->registerHandler('FormatHandler', $self, 'lazyinteger');
    $Processor->registerHandler('FormatHandler', $self, 'integer');
    $Processor->registerHandler('FormatHandler', $self, 'floor');
    $Processor->registerHandler('FormatHandler', $self, 'float');
    $Processor->registerHandler('FormatHandler', $self, 'money');
    $Processor->registerHandler('FormatHandler', $self, 'neg');
    $Processor->registerHandler('FormatHandler', $self, 'px');
    $Processor->registerHandler('FormatHandler', $self, 'number');
    $Processor->registerHandler('FormatHandler', $self, 'trimzeros');
    $Processor->registerHandler('VariableHandler', $self, 'ExampleFloat');
    $Processor->registerHandler('VariableHandler', $self, 'SmallExampleFloat');
    return 1;
}

sub Format {
    my $self = shift;
    my $Processor = shift;
    my ($Format, $Value, $TLEName) = @_;

    return undef unless defined $Value;
    if ($Format eq 'lazyinteger') {
        $Value =~ /[-]?\d+/;
        return $&;
    }

    if ($Format eq 'number') {
        $Value =~ /^\d+$/;
        return $&;
    }

    if ($Format eq 'trimzeros') {
        $Value =~ s/^0+//;
        return $Value;
    }
    my $Locale =  $Processor->tle('INPUT.Locale');
    # correct DE Posix
    my $Formatter = (defined $Locale && $Locale eq 'de_DE')
        ? DE_EPAGES::Core::API::Object::NumberFormat->new(
            mon_decimal_point  => ',',
            mon_thousands_sep  => '.',
            decimal_point      => ',',
            thousands_sep      => '.',
            int_curr_symbol => '',
            decimal_fill => 1,
            decimal_digits => 2,
            curr_pos_format => 'x ',
            curr_neg_format => '-x ',
            neg_format => '-x',
            int_currencyid => 'EUR',
        )
        : DE_EPAGES::Core::API::Object::NumberFormat->new()

    ;
    return $Formatter->format_price($Value) if $Format eq 'money';
    return 0.00 - $Value if $Format eq 'neg';
    if ($Format eq 'px'){
        return $Value.'px' if $Value =~ /^[+-]?\d+$/;
        return $Value;
    }
    return $Formatter->format_number($Value, 0) if $Format eq 'integer';
    return int($Value) if $Format eq 'floor';
    return $Formatter->format_number($Value);
}

sub tle {
    my $self = shift;
    my ($Processor, $TLEName) = @_;

    return 12.34 if $TLEName eq 'SmallExampleFloat' ;
    return 12345.6789 if $TLEName eq 'ExampleFloat' ;
}

sub existsTLE {
    my $self = shift;
    return 1;
}

1;

