#==============================================================================
# package      DE_EPAGES::WebService::API::WebService::BaseService
# state        public
# description  Base class for web services with permission handling.
#==============================================================================
package DE_EPAGES::WebService::API::WebService::BaseService;

use strict;
use Scalar::Util qw (blessed);
use MIME::Parser;
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::File qw ( ConvertFileNameChars );
use DE_EPAGES::Core::API::Log qw( GetLog LogDebug );
use DE_EPAGES::Core::API::String qw ( QuoteXML );
use DE_EPAGES::Core::API::Password  qw ( CheckPassword );
use DE_EPAGES::Core::API::PerlTools  qw (ElementOf IsA);
use DE_EPAGES::Core::API::TestParameter qw( TestObject );
use DE_EPAGES::Object::API::Factory qw( LoadRootObject ExistsObjectByPath LoadObjectByPath ExistsClassByAlias);
use DE_EPAGES::Object::API::Language qw( GetCodeByLanguageID GetPKeyLanguageByCode );
use DE_EPAGES::Permission::API::Permission qw( IsActionAllowed );
use DE_EPAGES::WebService::Servlet::Soap qw( GetSoapServer );
use DE_EPAGES::XML::API::SOAPDateTimeFormatXML;
use DE_EPAGES::WebInterface::API::ApplicationServer qw (GetAppServer);

#ERRORDEFINITION
sub ErrorDefinition {
    return {
        'MANDATORY_PARAMETER_MISSING' => 'Mandatory parameter #Parameter was not supplied.',
        'LANGUAGE_NOT_IN_STORE'       => 'Language #LanguageCode is not available for current store.',
        'ATTRIBUTE_DOES_NOT_EXIST'    => 'The attribute "#Attribute" does not exist in class "#Class".',
    }
}
#ERRORDEFINITION END

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Service = DE_EPAGES::WebService::API::WebService::BaseService->new( %Options );
#----------------------------------------------------------------------------------------
# description  Creates a new web service object as a blessed hash reference and adds
#               the values from the %Options hash.
#----------------------------------------------------------------------------------------
# input        %Options | user-defined values. Possible keys: <ul>
#                 <li>Object - site object to operate webservice on - object</li>
#               </ul>
#               | hash
# return       $Service | web service object | object
#========================================================================================
sub new {
    my $class = shift;
    my %Options = @_;
    $Options{'Object'} //= LoadRootObject();
    return bless {
        'XmlSchemaUrn' => 'http://www.w3.org/2001/XMLSchema',
        'EpagesTypesUrn' =>  'urn://epages.de/WebService/EpagesTypes/2005/01',
        'FormTypesUrn' =>  'urn://epages.de/WebService/FormTypes/2014/02',
        'UseGuidAsPath' => 0,
        %Options,
    }, $class;
}

#========================================================================================
# function     getXmlSchemaUrn
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Service->getXmlSchemaUrn();
#----------------------------------------------------------------------------------------
# description  returns the schema urn of the service (was set by constructor new()
#----------------------------------------------------------------------------------------
# return       $SchemaUrn | web service schema urn | string
#========================================================================================
sub getXmlSchemaUrn {
    return shift->{'XmlSchemaUrn'};
}

#========================================================================================
# function     getEpagesTypesUrn
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Service->getEpagesTypesUrn();
#----------------------------------------------------------------------------------------
# description  returns the epages types urn of the service (was set by constructor new()
#----------------------------------------------------------------------------------------
# return       $EpagesTypesUrn | web service epages types urn | string
#========================================================================================
sub getEpagesTypesUrn {
    return shift->{'EpagesTypesUrn'};
}

#========================================================================================
# function     getFormTypesUrn
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Service->getFormTypesUrn();
#----------------------------------------------------------------------------------------
# description  returns the form types urn of the service (was set by constructor new()
#----------------------------------------------------------------------------------------
# return       $FormTypesUrn | web service form types urn | string
#========================================================================================
sub getFormTypesUrn {
    return shift->{'FormTypesUrn'};
}

#========================================================================================
# function     CheckPermission
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IsAllowed = $self->CheckPermission( $WebService, $Method, $Login, $Password );
#----------------------------------------------------------------------------------------
# description  Checks if the user identified by $Login and $Password has permission to
#               to execute the method $Method of the web service $WebService.
#----------------------------------------------------------------------------------------
# input        $WebService | web service object | object
# input        $Method | web service method object | object
# input        $Login | login | string
# input        $Password | password | string
# return       $IsAllowed | true if the method call is allowed | boolean
#========================================================================================
sub CheckPermission {
    my $self = shift;
    my ($WebService, $Method, $Login, $Password) = @_;
    TestObject('WebService', $WebService);
    TestObject('Method', $Method);

    GetLog()->debug( sprintf 'CheckPermission(%s, %s, %s, %s)',
        $WebService->get('Alias'),
        $Method->get('Alias'),
        defined $Login ? $Login : 'undef',
        defined $Password ? $Password : 'undef'
    );

    my $User;
    if( defined $Login ) {

        unless (ExistsObjectByPath($Login)) {
            GetLog()->debug("User not found: $Login");

            # access denied for unknown login
            return 0;
        }

        $User = LoadObjectByPath( $Login );

        # check the password
        my $UserPassword = $User->get('Password');
        unless (CheckPassword($Password, $UserPassword)) {
            GetLog()->debug("Invalid password supplied for user: $Login ($Password ne $UserPassword)");
            return 0;
        }
    }

    $self->user($User);

    GetLog()->debug('User and Password correct.');

    my $isAllowed = IsActionAllowed( $self->site, $User, $Method );

    GetLog()->debug($isAllowed?'Action Allowed.':'Action NOT allowed.');

    return $isAllowed;
}

#========================================================================================
# function     user
# state        public
#----------------------------------------------------------------------------------------
# syntax       $User = $service->user;
# syntax       $User = $service->user($User);
#----------------------------------------------------------------------------------------
# description  Sets and returns the current user of the WebService.
#----------------------------------------------------------------------------------------
# input        $User | WebService user. (optional) Use it, to set the user. | object
# return       $User | WebService user | object
#========================================================================================
sub user {
    my $self = shift;
    my ($User) = @_;
    if (defined $User) {
        $self->{'User'} = $User;
    }
    return $self->{'User'};
}

#========================================================================================
# function     object
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object = $service->object;
# syntax       $Object = $service->object($Object);
#----------------------------------------------------------------------------------------
# description  Sets and returns the base object on which the WebService will act.
#               This is expected to be a Site object.
#               DEPRECATED: Use $service->site
#----------------------------------------------------------------------------------------
# input        $Object | base WebService object (optional). Use to set the base
#                         Webservice object | object
# return       $Object | base WebService object | object
#========================================================================================
sub object {
    my $self = shift;
    my ($Object) = @_;
    if (defined $Object) {
        $self->{'Object'} = $Object;
    }
    return $self->{'Object'};
}

#========================================================================================
# function     site
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Site = $service->site;
# syntax       $Site = $service->site($Site);
#----------------------------------------------------------------------------------------
# description  Sets and returns the base object on which the WebService will act.
#               This is expected to be a Site object.
#----------------------------------------------------------------------------------------
# input        $Site | base WebService object (optional). Use to set the base
#                         Webservice object | object
# return       $Site | base WebService object | object
#========================================================================================
sub site {
    my $self = shift;
    my ($Site) = @_;
    return $self->object($Site);
}

#========================================================================================
# function     mapType
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->mapType($aTypes, $Uri);
# example      $self->mapType(['TProduct, TPrice], 'urn://myservice/2005/06');
#----------------------------------------------------------------------------------------
# description  Map the given XSD Types to a specific Uri.
#----------------------------------------------------------------------------------------
# input        $aTypes | list of types | ref.array.string
# input        $Uri    | uri to map to | string
#========================================================================================
sub mapType {
    my $self = shift;
    my ($aTypes, $Uri) = @_;

    my $serializer = GetSoapServer()->serializer;
    $serializer->maptype->{$_} = $Uri foreach @$aTypes;
}

#========================================================================================
# function     soapError
# state        public
#----------------------------------------------------------------------------------------
# syntax       $SoapError = $service->soapError($Error);
# example      $Error = GetError();
#               $soapProduct = $self->soapDataType('TGetInfo_Return', {
#                  'Path'  => $self->soapPath($Product),
#                  'Error' => $self->soapError($Error)
#               });
#----------------------------------------------------------------------------------------
# description  Converts an error to a SOAP object
#----------------------------------------------------------------------------------------
# input        $Error | error object | DE_EPAGES::Core::API::Error
# return       $SoapError | SOAP object | object
#========================================================================================
sub soapError {
    my $self = shift;
    my ($Error) = @_;
    LogDebug('Error:', $Error);

    $self->mapType(['TError'], $self->getEpagesTypesUrn);
    return $self->soapDataType('TError', {'Message' => $self->soapString($Error->shortMessage)} );
}

#========================================================================================
# function     soapFormErrors
# state        private
#----------------------------------------------------------------------------------------
# syntax       $soapData = $self->soapFormErrors( $ahFormErrors )
#----------------------------------------------------------------------------------------
# description  No WebService-Method!
#               build a SOAP::Data structure of form errors
#----------------------------------------------------------------------------------------
# input        $ahFormErrors | list of attributes, with this keys
#                             <ul>
#                               <li>Code  - error code - string
#                               <li>Name  - name of responding field - string
#                               <li>Value - value of responding field - string
#                             </ul> | ref.array.hash
# return       $aSoapFormErrors | array of TFormError objects | ref.array
#========================================================================================
sub soapFormErrors {
    my $self = shift;
    my ($ahFormErrors) = @_;

    return undef unless scalar($ahFormErrors) > 0;

    my $aSoapFormErrors = [];
    foreach my $hFormError (@$ahFormErrors) {
        push @$aSoapFormErrors, SOAP::Data->type('TFormError')->name('Code' => {
            'Code'  => $self->soapString($hFormError->{'Reason'}),
            'FieldName'  => $self->soapString($hFormError->{'Name'}),
            'FieldValue'  => $self->soapString($hFormError->{'Value'})
        });
    }

    $self->mapType(['TFormError'], $self->getFormTypesUrn);

    return $aSoapFormErrors;
}

#========================================================================================
# function     soapDataValue
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapDataValue($Value, $Type);
# example      $self->soapDataValue('String', 'example');
# example      $self->soapDataValue('TMyType', $soapStructure );
# example      $self->soapDataValue('anyType', $hStructure );
# example      $self->soapDataValue('String', 'additionalAttr')->attr($hAttr);
#----------------------------------------------------------------------------------------
# description  build a SOAP::Data structure of passed type and value
#               the returned SOAP::Data object can call SOAP::Data methods like attr()
#----------------------------------------------------------------------------------------
# input        $Value | scalar value by simple type or soap data structure | string/hash/object
# input        $Type | simple, complex or new defined type
#                       simple types: string,float,integer,boolean, etc.
#                       complex types: array,anyType,enum
#                       new defined type: TError, TListOfPrices, etc. | string
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapDataValue { ## no critic qw( RequireArgUnpacking RequireCommentForExportedFunctions)
    return SOAP::Data->value($_[1])->type($_[2]);
}

#========================================================================================
# function     soapDataType
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapDataType($Type, $Value);
# example      $self->soapDataType('String', 'example');
# example      $self->soapDataType('TMyType', $soapStructure );
# example      $self->soapDataType('anyType', $hStructure );
# example      $self->soapDataType('String', 'additionalAttr')->attr($hAttr);
#----------------------------------------------------------------------------------------
# description  build a SOAP::Data structure of passed type and value
#               the returned SOAP::Data object can call SOAP::Data methods like attr()
#----------------------------------------------------------------------------------------
# input        $Type | simple, complex or new defined type
#                       simple types: string,float,integer,boolean, etc.
#                       complex types: array,anyType,enum
#                       new defined type: TError, TListOfPrices, etc. | string
# input        $Value | scalar value by simple type or soap data structure | string/hash/object
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapDataType { ## no critic qw( RequireArgUnpacking RequireCommentForExportedFunctions)
    return SOAP::Data->type($_[1],$_[2]);
}

#========================================================================================
# function     soapData
# state        public
#----------------------------------------------------------------------------------------
# syntax       soapData( $name, $value, $type );
#----------------------------------------------------------------------------------------
# description  Crates a simple SOAP::Data element containing cdata
#               DateTime object are auto formatted
#               value are quoted before SOAP::Lite do it automatically if no type given
#----------------------------------------------------------------------------------------
# input        $name | Element name | string
# input        $value | Element value | *
# input        $type | Element type (optional) | string
# return       $data | SOAP::Data element | object
#========================================================================================
sub soapData {
    my $self = shift;
    my ( $name, $value, $type ) = @_;

    if (IsA($value,'DateTime')) {
        return $self->soapDateTime($value)->name($name);
    } elsif ( $type ) {
        return $self->soapDataType($type,$value)->name($name);
    } else {
        $value = QuoteXML($value) unless ref $value;
        return SOAP::Data->name($name)->value($value);
    }
}

#========================================================================================
# function     soapString
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapString($Value);
# example      $self->soapString('example');
# example      $self->soapString('additionalAttr')->prefix('ns2');
#----------------------------------------------------------------------------------------
# description  build a SOAP::Data structure of type string
#               the returned SOAP::Data object can call SOAP::Data methods
#               like prefix() if need a specific name space
#----------------------------------------------------------------------------------------
# input        $Value | string value | string
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapString { ## no critic qw( RequireArgUnpacking RequireCommentForExportedFunctions)
    return SOAP::Data->type('string', $_[1]);
}

#========================================================================================
# function     soapFloat
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapFloat($Value);
# example      $self->soapFloat(17.559);
# example      $self->soapFloat(.001)->method();
#----------------------------------------------------------------------------------------
# description  build a SOAP::Data structure of type float
#               the returned SOAP::Data object can call SOAP::Data methods
#----------------------------------------------------------------------------------------
# input        $Value | float value | float
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapFloat { ## no critic qw( RequireArgUnpacking RequireCommentForExportedFunctions)
    return SOAP::Data->type('float', $_[1]);
}

#========================================================================================
# function     soapInt
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapInt($Value);
# example      $self->soapInt(17);
# example      $self->soapInt(1)->method();
#----------------------------------------------------------------------------------------
# description  build a SOAP::Data structure of type integer
#               the returned SOAP::Data object can call SOAP::Data methods
#----------------------------------------------------------------------------------------
# input        $Value | integer value | integer
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapInt { ## no critic qw( RequireArgUnpacking RequireCommentForExportedFunctions)
    return SOAP::Data->type('integer', $_[1]);
}

#========================================================================================
# function     soapBool
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapBool($isCondition);
# example      $self->soapBool(0);
# example      $self->soapBool(1)->method();
#----------------------------------------------------------------------------------------
# description  build a SOAP::Data structure of type boolean
#               the returned SOAP::Data object can call SOAP::Data methods
#----------------------------------------------------------------------------------------
# input        $Value | boolean value | boolean
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapBool { ## no critic qw( RequireArgUnpacking RequireCommentForExportedFunctions)
    return SOAP::Data->type('boolean', $_[1]);
}

#========================================================================================
# function     soapPath
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapPath($Object);
# example      $self->soapPath(0);
# example      $self->soapPath(1)->method();
#----------------------------------------------------------------------------------------
# description  build a SOAP::Data structure of type string
#               with object path value or undef if no path
#               the returned SOAP::Data object can call SOAP::Data methods
#----------------------------------------------------------------------------------------
# input        $Object | object | object
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapPath { ## no critic qw( RequireArgUnpacking RequireCommentForExportedFunctions)
    return $_[0]->soapString($_[1] ? $_[0]->_guidOrPath($_[1]) : undef );
}

#========================================================================================
# function     _guidOrPath
# state        private
#----------------------------------------------------------------------------------------
# syntax       $service->_guidOrPath($Object);
#----------------------------------------------------------------------------------------
# description  returns a guid or a pathString of the given defined object.
#----------------------------------------------------------------------------------------
# input        $Object | object (mandatory) | object
# return       $pathOrString | guid or path string | string
#========================================================================================
sub _guidOrPath { ## no critic qw( RequireArgUnpacking RequireCommentForExportedFunctions)
    return $_[0]->useGuidAsPath() ? $_[1]->guid : $_[1]->pathString;
}

#========================================================================================
# function     soapList
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapList($aList, $Type, $Uri);
# example      $self->soapList($aPaths,'string');
# example      $self->soapList($aAttributes,'TAttribute');
# example      $self->soapList($aVars,'TMyType','urn://vendor.do.main/WS/Types');
#----------------------------------------------------------------------------------------
# description  build a complex SOAP::Data structure of list (array) of same type
#               the array will passed with content to handle by SOAP::Lite
#               but empty array will handled with type and uri
#               the returned SOAP::Data object can call SOAP::Data methods
#----------------------------------------------------------------------------------------
# input        $aList | array of elements | ref.array
# input        $Type | type of an element | string
# input        $Uri | schema uri of type namespace | string
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapList {
    my $self = shift;
    my ($aList, $Type, $Uri) = @_;

    # return ref to list if it contains items to handle by SOAP::Lite encoder
    return $aList if scalar @$aList;

    if (not $Uri) {
        if ($Type eq 'TAttribute' or $Type eq 'TLocalizedValue') {
            $Uri = $self->getEpagesTypesUrn;
        }
        elsif ($Type eq 'TFormError') {
            $Uri = $self->getFormTypesUrn;
        }
        elsif (ElementOf($Type, qw (string integer float boolean)) ) {
            $Uri = $self->getXmlSchemaUrn ;
        }
        else {
            my $maptype = GetSoapServer()->serializer->maptype;
            if ( exists $maptype->{$Type} ) {
                $Uri = $maptype->{$Type}
            } else {
                Error('NO_NAMESPACE_FOR_TYPE', {'Type' => $Type});
            }
        }
    }

    # return empty soap encoded list (with correct set namespace and type
    my $soapenc = SOAP::Lite->encprefix();
    return SOAP::Data->value([])
                     ->attr({$soapenc.':arrayType' => 'ns:'.$Type.'[0]'})
                     ->uri( $Uri )
                     ->prefix('ns')
}

#========================================================================================
# function     soapLocList
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapLocList($object, $AttrName, $aLanguages, $Uri);
# example      $self->soapLocList($product, 'Description', ['de','en']);
#----------------------------------------------------------------------------------------
# description  build a complex SOAP::Data structure of list (array) of localized string
#               localized values of the passed attribute will get by soapLocalizedValues
#----------------------------------------------------------------------------------------
# input        $object | object | object
# input        $aList | array of elements | ref.array
# input        $Type | type of an element | string
# input        $Uri | schema uri of type namespace | string
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapLocList {
    my $self = shift;
    my ($object, $AttrName, $aLanguages, $Uri) = @_;
    return $self->soapList($self->soapLocalizedValues($object, $AttrName, $aLanguages ), 'TLocalizedValue', $Uri );
}

#========================================================================================
# function     soapDateTime
# state        public
#----------------------------------------------------------------------------------------
# syntax       $service->soapDateTime($datetime);
# example      $self->soapDateTime($Order->get('ShippedOn');
#----------------------------------------------------------------------------------------
# description  build a SOAP::Data structure of type datetime
#               datime will adjust to system time zone
#               and format by SOAPDateTimeFormatXML formatter
#               the returned SOAP::Data object can call SOAP::Data methods
#----------------------------------------------------------------------------------------
# input        $datetime | datetime value | datetime
# return       $soapDataType | SOAP::Data object | object
#========================================================================================
sub soapDateTime {
    my $self = shift;
    my ($datetime) = @_;

    my $dateparser = DE_EPAGES::XML::API::SOAPDateTimeFormatXML->new();
    my $SystemTimeZone = LoadRootObject()->get('TimeZone');

    my $datetimestr;
    if ($datetime) {
        $datetimestr = $dateparser->format_datetime($datetime, $SystemTimeZone);
    } else {
        $datetimestr = undef;
    }

    return SOAP::Data->value( $datetimestr )
                     ->attr({'xsi:type' => 'xsd:dateTime'});
}

sub soapAttributeValues {
    my $self = shift;
    my ($object, $aAttributes, $aLanguages) = @_;

    my @SoapAttributes = ();

    return [] unless $aAttributes;

    my $SystemTimeZone;
    foreach my $attributeId (@$aAttributes) {
        Error('ATTRIBUTE_DOES_NOT_EXIST', {
            'Attribute' => $attributeId,
            'Class' => $object->class->alias,
        }) unless $object->class->existsAttribute($attributeId);

        my $attribute = $object->class->attribute($attributeId);
        my $value;
        my $hAttribute = {
            'Name'  => $self->soapString( $attributeId ),
            'Type'  => $self->soapString( $attribute->get('Type') ),
        };

        if ($attribute->get('Type') eq 'LocalizedString' || $attribute->get('Type') eq 'LocalizedFile') {
            my $aLocValues = $self->soapLocalizedValues($object, $attributeId, $aLanguages );
            $hAttribute->{'LocalizedValues'} = $self->soapList($aLocValues, 'TLocalizedValue' );
        }
        else {
            $value = $object->get( $attributeId );

            #workaround to hande multi pre defs EPG-7293
            if ($attribute->get('Type') =~ /^PreDefMulti.*String$/) {
                $value = join';', map {$self->_guidOrPath($_)} @$value;
            }
            elsif (blessed($value)) {
                # if value is an object
                if ($value->isa('DE_EPAGES::Object::API::Object::Object')) {
                    #epages object
                    $value = $self->_guidOrPath($value);
                }
                elsif ($value->isa('DateTime')) {
                    #DateTime object
                    $SystemTimeZone = LoadRootObject()->get('TimeZone') unless defined $SystemTimeZone;
                    my $parser = DE_EPAGES::XML::API::SOAPDateTimeFormatXML->new();
                    $value = $parser->format_datetime($value, $SystemTimeZone);
                }
                else {
                    # ???
                    $value = blessed($value);
                }
            }
            elsif (ref $value) {
                # if value is a hash or array ref, return atleast the type
                $value = ref $value;
            }
            $hAttribute->{'Value'} = $self->soapString( $value );
        }

        push @SoapAttributes, SOAP::Data->type('TAttribute')->name('Attribute' => $hAttribute);
    }

    $self->mapType([qw (TAttribute TLocalizedValue)], $self->getEpagesTypesUrn);

    return \@SoapAttributes;
}

#========================================================================================
# function     soapLocalizedValues
# state        private
#----------------------------------------------------------------------------------------
# syntax       $soapData = $self->soapLocalizedValues( $object, $attributeId, $aLanguageIds )
#----------------------------------------------------------------------------------------
# description  No WebService-Method!
#               Gets values of an localized value.
#----------------------------------------------------------------------------------------
# input        $object | object to update | object
# input        $attributeId | attribute alias | string
# input        $aLanguageIds | List of LanguageIds | ref.array.string
# return       $soapData | SOAP::Data object | object
#========================================================================================
sub soapLocalizedValues {
    my $self = shift;
    my ($object, $attributeId, $aLanguageIds) = @_;

    my $aLocValues = [];
    foreach my $langid (@$aLanguageIds) {
        push @$aLocValues, SOAP::Data->type('TLocalizedValue')->name('LocalizedValue' => {
            'LanguageCode' => $self->soapString( GetCodeByLanguageID($langid) ),
            'Value'        => $self->soapString( $object->get( $attributeId, $langid ) ),
        });
    }
    return $aLocValues;
}

#========================================================================================
# function     setAttributeValues
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->setAttributeValues( $object, $ahAttributes )
#----------------------------------------------------------------------------------------
# description  No WebService-Method!
#               Sets attributes of a object.
#----------------------------------------------------------------------------------------
# input        $object | object to update | object
# input        $ahAttributes | list of attributes, with this keys
#                             <ul>
#                               <li>Name  - attribute id - string
#                               <li>Value - attribute value to set - string
#                             </ul> | ref.array.hash
#========================================================================================
sub setAttributeValues {
    my $self = shift;
    my ($object, $ahAttributes) = @_;

    return undef unless $ahAttributes;

    my $SystemTimeZone;
    foreach my $hAttribute (@$ahAttributes) {
        my $attributeId = $hAttribute->{'Name'};

        Error('ATTRIBUTE_DOES_NOT_EXIST', {
            'Attribute' => $attributeId,
            'Class' => $object->class->alias,
        }) unless $object->class->existsAttribute($attributeId);

        my $attribute = $object->class->attribute($attributeId);
        my $type = $attribute->get('Type');

        if ($type eq 'LocalizedString' || $type eq 'LocalizedFile') {
            $self->setLocalizedValues($object, $attributeId, $hAttribute->{'LocalizedValues'} );
        }
        #workaround to hande multi pre defs EPG-7293
        elsif ($type =~ /^PreDefMulti.*String$/) {
            my @values = map {LoadObjectByPath($_)} split/\;/, $hAttribute->{'Value'};
            $object->set({ $attributeId => \@values });
        }
        elsif (ExistsClassByAlias($type)) {
            my $value = defined $hAttribute->{'Value'}
                ? LoadObjectByPath( $hAttribute->{'Value'} )
                : undef;
            $object->set({ $attributeId => $value });
        }
        elsif ($type eq 'DateTime' or $type eq 'Date' or $type eq 'Time') {
            if (defined $hAttribute->{'Value'} ) {
                $SystemTimeZone = LoadRootObject()->get('TimeZone') unless defined $SystemTimeZone;
                my $parser = DE_EPAGES::XML::API::SOAPDateTimeFormatXML->new();
                my $value = $parser->parse_datetime($hAttribute->{'Value'}, $SystemTimeZone);
                $object->set({ $attributeId => $value });
            } else {
                $object->set({ $attributeId => undef });
            }
        }
        else { # string, integer, ...
            $object->set({$attributeId => $hAttribute->{'Value'}});
        }
    }
}

#========================================================================================
# function     setLocalizedValues
# state        private
#----------------------------------------------------------------------------------------
# syntax       $soapData = $self->setLocalizedValues( $object, $attributeId, $aValues )
#----------------------------------------------------------------------------------------
# description  No WebService-Method!
#               Sets values of an localized value.
#----------------------------------------------------------------------------------------
# input        $object | object to update | object
# input        $attributeId | attribute id | string
# input        $aValues | list of localized values, with this keys
#                           <ul>
#                             <li>LanguageCode  - language code (e.g. 'en') - string
#                             <li>Value - value in the according language - string
#                           </ul> | ref.array.hash
# return       $soapData | SOAP::Data object | object
#========================================================================================
sub setLocalizedValues {
    my $self = shift;
    my ($object, $attributeId, $aValues) = @_;

    foreach my $hLocValue (@$aValues) {
        my $langid = GetPKeyLanguageByCode($hLocValue->{'LanguageCode'});
        $object->set({ $attributeId => $hLocValue->{'Value'} }, $langid);
    }
    return 1;
}

#========================================================================================
# function     checkMandatoryFields
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->checkMandatoryFields( $hStructure, $aField )
#----------------------------------------------------------------------------------------
# description  No WebService-Method!
#               checks if fields are in structure and not empty
#               Throws the error MANDATORY_PARAMETER_MISSING if any data is missing.
#----------------------------------------------------------------------------------------
# input        $hStructure | hash of object data | ref.hash.*
# input        $aField | field names to check | ref.array.string
#========================================================================================
sub checkMandatoryFields {
    my $self = shift;
    my ($hStructure, $aFields) = @_;

    foreach my $field (@$aFields) {
        if (not defined $hStructure->{$field} or $hStructure->{$field} !~ /\S/) {
            Error('MANDATORY_PARAMETER_MISSING', {'Parameter' => $field});
        }
    }
    return;
}

#========================================================================================
# function     getLanguageIDs
# state        private
#----------------------------------------------------------------------------------------
# syntax       $self->getLanguageIDs($aLanguageCodes )
#----------------------------------------------------------------------------------------
# description  No WebService-Method!
#               gets the language ids of a shop. result set can be restricted by a
#               list of language codes.
#               if language codes are omitted, all shop language ids will be returned.
#----------------------------------------------------------------------------------------
# input        $aLanguageCodes | list of language codes,
#                                 subset of shop languages (optional) | ref.array.string
# return       $aLanguageIDs | list of shop language ids (or subset) | ref.array.int
#========================================================================================
sub getLanguageIDs {
    my $self = shift;
    my ($aLanguageCodes) = @_;

    my @LanguageIDs = ();

    my @StoreLanguages = map {$_->{'LanguageID'}} @{LoadRootObject()->get('Languages')};
    return \@StoreLanguages unless scalar(@$aLanguageCodes);

    foreach my $langcode (@$aLanguageCodes) {
        my $langid = GetPKeyLanguageByCode(lc $langcode);
        if (not ElementOf($langid, @StoreLanguages)) {
            Error('LANGUAGE_NOT_IN_STORE', {'LanguageID' => $langid, 'LanguageCode' => $langcode});
        }
        push(@LanguageIDs, $langid);
    }
    return \@LanguageIDs;
}


#========================================================================================
# function     getMimeParts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->getMimeParts()
#----------------------------------------------------------------------------------------
# description  get the MIME::Entity parts of the current request.
#               All parts are returned as a list of MIME::Entity objects
#               Used to transfer Binaries via Webservices
#               (If you would get the attachment of a part you have to filter out the
#               MIME::Entity parts with attachments!)
#----------------------------------------------------------------------------------------
# return       $aEntities | list of MIME::Entity objects | ref.array.object
#========================================================================================
sub getMimeParts {
    my $self = shift;

    my $Request = GetAppServer()->currentRequest();
    my $parser = MIME::Parser->new();
    my $TopEntity = $parser->parse_data($Request->headers->as_string.$Request->content);

    my @aEntities;
    for (my $i = 0; $i <= $TopEntity->parts()-1; $i++ ) {
        push @aEntities,$TopEntity->parts($i);
    }
    return \@aEntities;
}

#========================================================================================
# function     useGuidAsPath
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Service->useGuidAsPath();
#----------------------------------------------------------------------------------------
# description  returns true, if object GUIDs should be used instead of object paths.
#               This can be passed as constructor argument.
#----------------------------------------------------------------------------------------
# return       $UseGuidAsPath | true, if set, false otherwise | string
#========================================================================================
sub useGuidAsPath {
    return shift->{'UseGuidAsPath'};
}


# New Package starts here ================================================================

# Serialize DateTime objects as xsd:dateTime
# Note: xsd:dateTime is not defined in XML Schema 1999
package SOAP::XMLSchema2001::Serializer; ##no critic (ProhibitMultiplePackages)
sub as_DateTime { ##no critic (RequireArgUnpacking)
    return SOAP::XMLSchema2001::Serializer::as_dateTime(@_);
}


1;
